import datetime
import json
import unittest
import urllib.request
from unittest.mock import patch
from urllib.error import HTTPError, URLError

import responses
from django import template
from django.core.exceptions import ValidationError
from django.test import TestCase, override_settings
from django.urls import reverse
from django.utils.timezone import make_aware, now
from responses import matchers

from wagtail import blocks
from wagtail.embeds import oembed_providers
from wagtail.embeds.blocks import EmbedBlock, EmbedValue
from wagtail.embeds.embeds import get_embed, get_embed_hash
from wagtail.embeds.exceptions import (
    EmbedNotFoundException,
    EmbedUnsupportedProviderException,
)
from wagtail.embeds.finders import get_finders
from wagtail.embeds.finders.embedly import (
    AccessDeniedEmbedlyException,
    EmbedlyException,
)
from wagtail.embeds.finders.embedly import EmbedlyFinder as EmbedlyFinder
from wagtail.embeds.finders.facebook import AccessDeniedFacebookOEmbedException
from wagtail.embeds.finders.facebook import FacebookOEmbedFinder as FacebookOEmbedFinder
from wagtail.embeds.finders.instagram import AccessDeniedInstagramOEmbedException
from wagtail.embeds.finders.instagram import (
    InstagramOEmbedFinder as InstagramOEmbedFinder,
)
from wagtail.embeds.finders.oembed import OEmbedFinder as OEmbedFinder
from wagtail.embeds.models import Embed
from wagtail.embeds.templatetags.wagtailembeds_tags import embed_tag
from wagtail.test.utils import WagtailTestUtils

try:
    import embedly  # noqa: F401

    no_embedly = False
except ImportError:
    no_embedly = True


class DummyFinder:
    def __init__(self, finder_func):
        self.finder_func = finder_func

    def accept(self, url):
        # Always accept for test purposes.
        return True

    def find_embed(self, url, max_width=None, **kwargs):
        return self.finder_func(url, max_width, **kwargs)


class TestGetFinders(TestCase):
    def test_defaults_to_oembed(self):
        finders = get_finders()

        self.assertEqual(len(finders), 1)
        self.assertIsInstance(finders[0], OEmbedFinder)

    # New WAGTAILEMBEDS_FINDERS setting

    @override_settings(
        WAGTAILEMBEDS_FINDERS=[{"class": "wagtail.embeds.finders.oembed"}]
    )
    def test_new_find_oembed(self):
        finders = get_finders()

        self.assertEqual(len(finders), 1)
        self.assertIsInstance(finders[0], OEmbedFinder)

    @override_settings(
        WAGTAILEMBEDS_FINDERS=[
            {
                "class": "wagtail.embeds.finders.embedly",
                "key": "foo",
            }
        ]
    )
    def test_new_find_embedly(self):
        finders = get_finders()

        self.assertEqual(len(finders), 1)
        self.assertIsInstance(finders[0], EmbedlyFinder)
        self.assertEqual(finders[0].get_key(), "foo")

    @override_settings(
        WAGTAILEMBEDS_FINDERS=[
            {"class": "wagtail.embeds.finders.oembed", "options": {"foo": "bar"}}
        ]
    )
    def test_new_find_oembed_with_options(self):
        finders = get_finders()

        self.assertEqual(len(finders), 1)
        self.assertIsInstance(finders[0], OEmbedFinder)
        self.assertEqual(finders[0].options, {"foo": "bar"})

    @override_settings(
        WAGTAILEMBEDS_FINDERS=[
            {
                "class": "wagtail.embeds.finders.instagram",
                "app_id": "1234567890",
                "app_secret": "abcdefghijklmnop",
            },
        ]
    )
    def test_find_instagram_oembed_with_options(self):
        finders = get_finders()

        self.assertEqual(len(finders), 1)
        self.assertIsInstance(finders[0], InstagramOEmbedFinder)
        self.assertEqual(finders[0].app_id, "1234567890")
        self.assertEqual(finders[0].app_secret, "abcdefghijklmnop")
        # omitscript defaults to False
        self.assertIs(finders[0].omitscript, False)

    @override_settings(
        WAGTAILEMBEDS_FINDERS=[
            {
                "class": "wagtail.embeds.finders.facebook",
                "app_id": "1234567890",
                "app_secret": "abcdefghijklmnop",
            },
        ]
    )
    def test_find_facebook_oembed_with_options(self):
        finders = get_finders()

        self.assertEqual(len(finders), 1)
        self.assertIsInstance(finders[0], FacebookOEmbedFinder)
        self.assertEqual(finders[0].app_id, "1234567890")
        self.assertEqual(finders[0].app_secret, "abcdefghijklmnop")
        # omitscript defaults to False
        self.assertIs(finders[0].omitscript, False)


class TestEmbeds(TestCase):
    def setUp(self):
        self.hit_count = 0

    def dummy_finder(self, url, max_width=None, max_height=None):
        # Up hit count
        self.hit_count += 1

        # Return a pretend record
        return {
            "title": "Test: " + url,
            "type": "video",
            "width": max_width if max_width else 640,
            "height": 480,
            "html": "<p>Blah blah blah</p>",
        }

    @override_settings(WAGTAILEMBEDS_RESPONSIVE_HTML=True)
    def test_get_embed_responsive(self):
        with patch("wagtail.embeds.embeds.get_finders") as get_finders_mock:
            # Configure get_finders to return our dummy finder properly
            dummy_finder = DummyFinder(self.dummy_finder)
            get_finders_mock.return_value = [dummy_finder]

            embed = get_embed("www.test.com/1234", max_width=400)

        # Check that the embed is correct
        self.assertEqual(embed.title, "Test: www.test.com/1234")
        self.assertEqual(embed.type, "video")
        self.assertEqual(embed.width, 400)
        self.assertEqual(embed.thumbnail_url, "")

        # Check ratio calculations
        self.assertEqual(embed.ratio, 480 / 400)
        self.assertEqual(embed.ratio_css, "120.0%")
        self.assertTrue(embed.is_responsive)

        # Check that there has only been one hit to the backend
        self.assertEqual(self.hit_count, 1)

        # Look for the same embed again and check the hit count hasn't increased
        with patch("wagtail.embeds.embeds.get_finders") as get_finders_mock:
            get_finders_mock.return_value = [DummyFinder(self.dummy_finder)]
            get_embed("www.test.com/1234", max_width=400)
        self.assertEqual(self.hit_count, 1)

        # Look for a different embed, hit count should increase
        with patch("wagtail.embeds.embeds.get_finders") as get_finders_mock:
            get_finders_mock.return_value = [DummyFinder(self.dummy_finder)]
            get_embed("www.test.com/4321", max_width=400)
        self.assertEqual(self.hit_count, 2)

        # Look for the same embed with a different width, this should also increase hit count
        with patch("wagtail.embeds.embeds.get_finders") as get_finders_mock:
            get_finders_mock.return_value = [DummyFinder(self.dummy_finder)]
            get_embed("www.test.com/4321")
        self.assertEqual(self.hit_count, 3)

    def test_get_embed_nonresponsive(self):
        with patch("wagtail.embeds.embeds.get_finders") as get_finders:
            get_finders.return_value = [DummyFinder(self.dummy_finder)]
            embed = get_embed("www.test.com/1234", max_width=400)

        # Check that the embed is correct
        self.assertEqual(embed.title, "Test: www.test.com/1234")
        self.assertEqual(embed.type, "video")
        self.assertEqual(embed.width, 400)
        self.assertFalse(embed.is_responsive)
        self.assertIsNone(embed.cache_until)

    def dummy_cache_until_finder(self, url, max_width=None, max_height=None):
        # Up hit count
        self.hit_count += 1

        # Return a pretend record
        return {
            "title": "Test: " + url,
            "type": "video",
            "width": max_width if max_width else 640,
            "height": 480,
            "html": "<p>Blah blah blah</p>",
            "cache_until": make_aware(datetime.datetime(2001, 2, 3)),
        }

    def test_get_embed_cache_until(self):
        # Patch get_finders to always return our dummy_cache_until_finder
        with patch("wagtail.embeds.embeds.get_finders") as get_finders:
            get_finders.return_value = [DummyFinder(self.dummy_cache_until_finder)]
            embed = get_embed("www.test.com/1234", max_width=400)
            self.assertEqual(
                embed.cache_until, make_aware(datetime.datetime(2001, 2, 3))
            )
            self.assertEqual(self.hit_count, 1)

            # expired cache_until should be ignored
            embed_2 = get_embed("www.test.com/1234", max_width=400)
            self.assertEqual(self.hit_count, 2)

            # future cache_until should not be ignored
            future_dt = now() + datetime.timedelta(minutes=1)
            embed.cache_until = future_dt
            embed.save()
            embed_3 = get_embed("www.test.com/1234", max_width=400)
            self.assertEqual(self.hit_count, 2)

            # ensure we've received the same embed
            self.assertEqual(embed, embed_2)
            self.assertEqual(embed, embed_3)
            self.assertEqual(embed_3.cache_until, future_dt)

    def dummy_finder_invalid_width(self, url, max_width=None, max_height=None):
        # Return a record with an invalid width
        return {
            "title": "Test: " + url,
            "type": "video",
            "thumbnail_url": "",
            "width": "100%",
            "height": 480,
            "html": "<p>Blah blah blah</p>",
        }

    def test_invalid_width(self):
        with patch("wagtail.embeds.embeds.get_finders") as get_finders:
            get_finders.return_value = [DummyFinder(self.dummy_finder_invalid_width)]
            embed = get_embed("www.test.com/1234", max_width=400)

        # Width must be set to None
        self.assertIsNone(embed.width)

    def test_no_html(self):
        def no_html_finder(url, max_width=None, max_height=None):
            """
            A finder which returns everything but HTML
            """
            embed = self.dummy_finder(url, max_width)
            embed["html"] = None
            return embed

        with patch("wagtail.embeds.embeds.get_finders") as get_finders:
            get_finders.return_value = [DummyFinder(no_html_finder)]
            embed = get_embed("www.test.com/1234", max_width=400)

        self.assertEqual(embed.html, "")

    @override_settings(WAGTAILEMBEDS_FINDERS=[])
    def test_no_finders_available(self):
        with self.assertRaises(EmbedUnsupportedProviderException):
            get_embed("www.test.com/1234", max_width=400)


class TestEmbedHash(TestCase):
    def test_get_embed_hash(self):
        url = "www.test.com/1234"
        self.assertEqual(get_embed_hash(url), "9a4cfc187266026cd68160b5db572629")
        self.assertEqual(get_embed_hash(url, 0), "946fb9597a6c74ab3cef1699eff7fde7")
        self.assertEqual(get_embed_hash(url, 1), "427830227a86093b50417e11dbd2f28e")


class TestChooser(WagtailTestUtils, TestCase):
    def setUp(self):
        # login
        self.login()

    def test_chooser(self):
        r = self.client.get("/admin/embeds/chooser/")
        self.assertEqual(r.status_code, 200)

    def test_chooser_with_edit_params(self):
        r = self.client.get("/admin/embeds/chooser/?url=http://example2.com")
        self.assertEqual(r.status_code, 200)
        response_json = json.loads(r.content.decode())
        self.assertEqual(response_json["step"], "chooser")
        self.assertIn('value="http://example2.com"', response_json["html"])

    @patch("wagtail.embeds.embeds.get_embed")
    def test_submit_valid_embed(self, get_embed):
        get_embed.return_value = Embed(
            html='<img src="http://www.example.com" />', title="An example embed"
        )

        response = self.client.post(
            reverse("wagtailembeds:chooser_upload"),
            {"embed-chooser-url": "http://www.example.com/"},
        )
        self.assertEqual(response.status_code, 200)
        response_json = json.loads(response.content.decode())
        self.assertEqual(response_json["step"], "embed_chosen")
        self.assertEqual(response_json["embed_data"]["title"], "An example embed")

    @patch("wagtail.embeds.embeds.get_embed")
    def test_submit_unrecognised_embed(self, get_embed):
        get_embed.side_effect = EmbedNotFoundException

        response = self.client.post(
            reverse("wagtailembeds:chooser_upload"),
            {"embed-chooser-url": "http://www.example.com/"},
        )
        self.assertEqual(response.status_code, 200)

        response_json = json.loads(response.content.decode())
        self.assertEqual(response_json["step"], "chooser")
        self.assertIn("Cannot find an embed for this URL.", response_json["html"])


class TestEmbedly(TestCase):
    @unittest.skipIf(no_embedly, "Embedly is not installed")
    def test_embedly_oembed_called_with_correct_arguments(self):
        with patch("embedly.Embedly.oembed") as oembed:
            oembed.return_value = {"type": "photo", "url": "http://www.example.com"}

            EmbedlyFinder(key="foo").find_embed("http://www.example.com")
            oembed.assert_called_with("http://www.example.com", better=False)

            EmbedlyFinder(key="foo").find_embed("http://www.example.com", max_width=100)
            oembed.assert_called_with(
                "http://www.example.com", maxwidth=100, better=False
            )

    @unittest.skipIf(no_embedly, "Embedly is not installed")
    def test_embedly_401(self):
        with patch("embedly.Embedly.oembed") as oembed:
            oembed.return_value = {
                "type": "photo",
                "url": "http://www.example.com",
                "error": True,
                "error_code": 401,
            }
            self.assertRaises(
                AccessDeniedEmbedlyException,
                EmbedlyFinder(key="foo").find_embed,
                "http://www.example.com",
            )

    @unittest.skipIf(no_embedly, "Embedly is not installed")
    def test_embedly_403(self):
        with patch("embedly.Embedly.oembed") as oembed:
            oembed.return_value = {
                "type": "photo",
                "url": "http://www.example.com",
                "error": True,
                "error_code": 403,
            }
            self.assertRaises(
                AccessDeniedEmbedlyException,
                EmbedlyFinder(key="foo").find_embed,
                "http://www.example.com",
            )

    @unittest.skipIf(no_embedly, "Embedly is not installed")
    def test_embedly_404(self):
        with patch("embedly.Embedly.oembed") as oembed:
            oembed.return_value = {
                "type": "photo",
                "url": "http://www.example.com",
                "error": True,
                "error_code": 404,
            }
            self.assertRaises(
                EmbedNotFoundException,
                EmbedlyFinder(key="foo").find_embed,
                "http://www.example.com",
            )

    @unittest.skipIf(no_embedly, "Embedly is not installed")
    def test_embedly_other_error(self):
        with patch("embedly.Embedly.oembed") as oembed:
            oembed.return_value = {
                "type": "photo",
                "url": "http://www.example.com",
                "error": True,
                "error_code": 999,
            }
            self.assertRaises(
                EmbedlyException,
                EmbedlyFinder(key="foo").find_embed,
                "http://www.example.com",
            )

    @unittest.skipIf(no_embedly, "Embedly is not installed")
    def test_embedly_html_conversion(self):
        with patch("embedly.Embedly.oembed") as oembed:
            oembed.return_value = {"type": "photo", "url": "http://www.example.com"}
            result = EmbedlyFinder(key="foo").find_embed("http://www.example.com")
            self.assertEqual(
                result["html"], '<img src="http://www.example.com" alt="">'
            )

            oembed.return_value = {"type": "something else", "html": "<foo>bar</foo>"}
            result = EmbedlyFinder(key="foo").find_embed("http://www.example.com")
            self.assertEqual(result["html"], "<foo>bar</foo>")

    @unittest.skipIf(no_embedly, "Embedly is not installed")
    def test_embedly_return_value(self):
        with patch("embedly.Embedly.oembed") as oembed:
            oembed.return_value = {"type": "something else", "html": "<foo>bar</foo>"}
            result = EmbedlyFinder(key="foo").find_embed("http://www.example.com")
            self.assertEqual(
                result,
                {
                    "title": "",
                    "author_name": "",
                    "provider_name": "",
                    "type": "something else",
                    "thumbnail_url": None,
                    "width": None,
                    "height": None,
                    "html": "<foo>bar</foo>",
                },
            )

            oembed.return_value = {
                "type": "something else",
                "author_name": "Alice",
                "provider_name": "Bob",
                "title": "foo",
                "thumbnail_url": "http://www.example.com",
                "width": 100,
                "height": 100,
                "html": "<foo>bar</foo>",
            }
            result = EmbedlyFinder(key="foo").find_embed("http://www.example.com")
            self.assertEqual(
                result,
                {
                    "type": "something else",
                    "author_name": "Alice",
                    "provider_name": "Bob",
                    "title": "foo",
                    "thumbnail_url": "http://www.example.com",
                    "width": 100,
                    "height": 100,
                    "html": "<foo>bar</foo>",
                },
            )


class TestOembed(TestCase):
    def test_oembed_invalid_provider(self):
        self.assertRaises(EmbedNotFoundException, OEmbedFinder().find_embed, "foo")

    @responses.activate
    def test_oembed_invalid_request(self):
        # no response set up, so responses will raise a ConnectionError
        self.assertRaises(
            EmbedNotFoundException,
            OEmbedFinder().find_embed,
            "https://www.youtube.com/watch/",
        )

    @responses.activate
    def test_oembed_non_json_response(self):
        responses.get(
            url="https://www.youtube.com/oembed",
            match=[
                matchers.query_param_matcher(
                    {
                        "url": "https://www.youtube.com/watch?v=ReblZ7o7lu4",
                        "format": "json",
                    }
                ),
            ],
            body="foo",
        )
        self.assertRaises(
            EmbedNotFoundException,
            OEmbedFinder().find_embed,
            "https://www.youtube.com/watch?v=ReblZ7o7lu4",
        )

    @responses.activate
    def test_oembed_photo_request(self):
        responses.get(
            url="https://www.youtube.com/oembed",
            match=[
                matchers.query_param_matcher(
                    {"url": "https://www.youtube.com/watch/", "format": "json"}
                ),
            ],
            json={"type": "photo", "url": "http://www.example.com"},
        )
        result = OEmbedFinder().find_embed("https://www.youtube.com/watch/")
        self.assertEqual(result["type"], "photo")
        self.assertEqual(result["html"], '<img src="http://www.example.com" alt="">')

    @responses.activate
    def test_oembed_return_values(self):
        responses.get(
            url="https://www.youtube.com/oembed",
            match=[
                matchers.query_param_matcher(
                    {"url": "https://www.youtube.com/watch/", "format": "json"}
                ),
            ],
            json={
                "type": "something",
                "url": "http://www.example.com",
                "title": "test_title",
                "author_name": "test_author",
                "provider_name": "test_provider_name",
                "thumbnail_url": "test_thumbail_url",
                "width": "test_width",
                "height": "test_height",
                "html": "test_html",
            },
        )
        result = OEmbedFinder().find_embed("https://www.youtube.com/watch/")
        self.assertEqual(
            result,
            {
                "type": "something",
                "title": "test_title",
                "author_name": "test_author",
                "provider_name": "test_provider_name",
                "thumbnail_url": "test_thumbail_url",
                "width": "test_width",
                "height": "test_height",
                "html": "test_html",
            },
        )

    @patch("django.utils.timezone.now")
    @responses.activate
    def test_oembed_cache_until(self, now):
        responses.get(
            url="https://www.youtube.com/oembed",
            match=[
                matchers.query_param_matcher(
                    {"url": "https://www.youtube.com/watch/", "format": "json"}
                ),
            ],
            json={
                "type": "something",
                "url": "http://www.example.com",
                "title": "test_title",
                "author_name": "test_author",
                "provider_name": "test_provider_name",
                "thumbnail_url": "test_thumbail_url",
                "width": "test_width",
                "height": "test_height",
                "html": "test_html",
                "cache_age": 3600,
            },
        )
        now.return_value = make_aware(datetime.datetime(2001, 2, 3))
        result = OEmbedFinder().find_embed("https://www.youtube.com/watch/")
        self.assertEqual(
            result,
            {
                "type": "something",
                "title": "test_title",
                "author_name": "test_author",
                "provider_name": "test_provider_name",
                "thumbnail_url": "test_thumbail_url",
                "width": "test_width",
                "height": "test_height",
                "html": "test_html",
                "cache_until": make_aware(datetime.datetime(2001, 2, 3, hour=1)),
            },
        )

    @patch("django.utils.timezone.now")
    @responses.activate
    def test_oembed_cache_until_as_string(self, now):
        responses.get(
            url="https://www.youtube.com/oembed",
            match=[
                matchers.query_param_matcher(
                    {"url": "https://www.youtube.com/watch/", "format": "json"}
                ),
            ],
            json={
                "type": "something",
                "url": "http://www.example.com",
                "title": "test_title",
                "author_name": "test_author",
                "provider_name": "test_provider_name",
                "thumbnail_url": "test_thumbail_url",
                "width": "test_width",
                "height": "test_height",
                "html": "test_html",
                "cache_age": "3600",
            },
        )
        now.return_value = make_aware(datetime.datetime(2001, 2, 3))
        result = OEmbedFinder().find_embed("https://www.youtube.com/watch/")
        self.assertEqual(
            result,
            {
                "type": "something",
                "title": "test_title",
                "author_name": "test_author",
                "provider_name": "test_provider_name",
                "thumbnail_url": "test_thumbail_url",
                "width": "test_width",
                "height": "test_height",
                "html": "test_html",
                "cache_until": make_aware(datetime.datetime(2001, 2, 3, hour=1)),
            },
        )

    def test_oembed_accepts_known_provider(self):
        finder = OEmbedFinder(providers=[oembed_providers.youtube])
        self.assertTrue(finder.accept("https://www.youtube.com/watch/"))

    def test_oembed_doesnt_accept_unknown_provider(self):
        finder = OEmbedFinder(providers=[oembed_providers.twitter])
        self.assertFalse(finder.accept("https://www.youtube.com/watch/"))

    @responses.activate
    def test_endpoint_with_format_param(self):
        responses.get(
            url="https://www.vimeo.com/api/oembed.json",
            match=[
                matchers.query_param_matcher(
                    {"url": "https://vimeo.com/217403396", "format": "json"}
                ),
            ],
            json={"type": "video", "url": "http://www.example.com"},
        )
        result = OEmbedFinder().find_embed("https://vimeo.com/217403396")
        self.assertEqual(result["type"], "video")


class TestInstagramOEmbed(TestCase):
    def setUp(self):
        class DummyResponse:
            def read(self):
                return b"""{
                    "type": "something",
                    "url": "http://www.example.com",
                    "title": "test_title",
                    "author_name": "test_author",
                    "provider_name": "Instagram",
                    "thumbnail_url": "test_thumbail_url",
                    "width": "test_width",
                    "height": "test_height",
                    "html": "<blockquote class=\\\"instagram-media\\\">Content</blockquote>"
                }"""

        self.dummy_response = DummyResponse()

    def test_instagram_oembed_only_accepts_new_url_patterns(self):
        finder = InstagramOEmbedFinder()
        self.assertTrue(
            finder.accept(
                "https://www.instagram.com/p/CHeRxmnDSYe/?utm_source=ig_embed"
            )
        )
        self.assertTrue(
            finder.accept(
                "https://www.instagram.com/tv/CZMkxGaIXk3/?utm_source=ig_embed"
            )
        )
        self.assertTrue(
            finder.accept(
                "https://www.instagram.com/reel/CZMs3O_I22w/?utm_source=ig_embed"
            )
        )
        self.assertFalse(
            finder.accept("https://instagr.am/p/CHeRxmnDSYe/?utm_source=ig_embed")
        )

    @patch("urllib.request.urlopen")
    def test_instagram_oembed_return_values(self, urlopen):
        urlopen.return_value = self.dummy_response
        result = InstagramOEmbedFinder(app_id="123", app_secret="abc").find_embed(
            "https://instagram.com/p/CHeRxmnDSYe/"
        )
        self.assertEqual(
            result,
            {
                "type": "something",
                "title": "test_title",
                "author_name": "test_author",
                "provider_name": "Instagram",
                "thumbnail_url": "test_thumbail_url",
                "width": "test_width",
                "height": "test_height",
                "html": '<blockquote class="instagram-media">Content</blockquote>',
            },
        )
        # check that a request was made with the expected URL / authentication
        request = urlopen.call_args[0][0]
        self.assertEqual(
            request.get_full_url(),
            "https://graph.facebook.com/v11.0/instagram_oembed?url=https%3A%2F%2Finstagram.com%2Fp%2FCHeRxmnDSYe%2F&format=json",
        )
        self.assertEqual(request.get_header("Authorization"), "Bearer 123|abc")

    def test_instagram_request_denied_401(self):
        err = HTTPError(
            "https://instagram.com/p/CHeRxmnDSYe/",
            code=401,
            msg="invalid credentials",
            hdrs={},
            fp=None,
        )
        config = {"side_effect": err}
        with patch.object(urllib.request, "urlopen", **config):
            self.assertRaises(
                AccessDeniedInstagramOEmbedException,
                InstagramOEmbedFinder().find_embed,
                "https://instagram.com/p/CHeRxmnDSYe/",
            )

    def test_instagram_request_not_found(self):
        err = HTTPError(
            "https://instagram.com/p/badrequest/",
            code=404,
            msg="Not Found",
            hdrs={},
            fp=None,
        )
        config = {"side_effect": err}
        with patch.object(urllib.request, "urlopen", **config):
            self.assertRaises(
                EmbedNotFoundException,
                InstagramOEmbedFinder().find_embed,
                "https://instagram.com/p/CHeRxmnDSYe/",
            )

    def test_instagram_failed_request(self):
        config = {"side_effect": URLError(reason="Testing error handling")}
        with patch.object(urllib.request, "urlopen", **config):
            self.assertRaises(
                EmbedNotFoundException,
                InstagramOEmbedFinder().find_embed,
                "https://instagram.com/p/CHeRxmnDSYe/",
            )


class TestFacebookOEmbed(TestCase):
    def setUp(self):
        class DummyResponse:
            def read(self):
                return b"""{
                    "type": "something",
                    "url": "http://www.example.com",
                    "title": "test_title",
                    "author_name": "test_author",
                    "provider_name": "Facebook",
                    "width": "test_width",
                    "height": "test_height",
                    "html": "<blockquote class=\\\"facebook-media\\\">Content</blockquote>"
                }"""

        self.dummy_response = DummyResponse()

    def test_facebook_oembed_accepts_various_url_patterns(self):
        finder = FacebookOEmbedFinder()
        self.assertTrue(
            finder.accept("https://www.facebook.com/testuser/posts/10157389310497085")
        )
        self.assertTrue(finder.accept("https://fb.watch/ABC123eew/"))

    @patch("urllib.request.urlopen")
    def test_facebook_oembed_return_values(self, urlopen):
        urlopen.return_value = self.dummy_response
        result = FacebookOEmbedFinder(app_id="123", app_secret="abc").find_embed(
            "https://fb.watch/ABC123eew/"
        )
        self.assertEqual(
            result,
            {
                "type": "something",
                "title": "test_title",
                "author_name": "test_author",
                "provider_name": "Facebook",
                "thumbnail_url": None,
                "width": "test_width",
                "height": "test_height",
                "html": '<blockquote class="facebook-media">Content</blockquote>',
            },
        )
        # check that a request was made with the expected URL / authentication
        request = urlopen.call_args[0][0]
        self.assertEqual(
            request.get_full_url(),
            "https://graph.facebook.com/v11.0/oembed_video?url=https%3A%2F%2Ffb.watch%2FABC123eew%2F&format=json",
        )
        self.assertEqual(request.get_header("Authorization"), "Bearer 123|abc")

    def test_facebook_request_denied_401(self):
        err = HTTPError(
            "https://fb.watch/ABC123eew/",
            code=401,
            msg="invalid credentials",
            hdrs={},
            fp=None,
        )
        config = {"side_effect": err}
        with patch.object(urllib.request, "urlopen", **config):
            self.assertRaises(
                AccessDeniedFacebookOEmbedException,
                FacebookOEmbedFinder().find_embed,
                "https://fb.watch/ABC123eew/",
            )

    def test_facebook_request_not_found(self):
        err = HTTPError(
            "https://fb.watch/ABC123eew/", code=404, msg="Not Found", hdrs={}, fp=None
        )
        config = {"side_effect": err}
        with patch.object(urllib.request, "urlopen", **config):
            self.assertRaises(
                EmbedNotFoundException,
                FacebookOEmbedFinder().find_embed,
                "https://fb.watch/ABC123eew/",
            )

    def test_facebook_failed_request(self):
        config = {"side_effect": URLError(reason="Testing error handling")}
        with patch.object(urllib.request, "urlopen", **config):
            self.assertRaises(
                EmbedNotFoundException,
                FacebookOEmbedFinder().find_embed,
                "https://fb.watch/ABC123eew/",
            )


class TestEmbedTag(TestCase):
    @patch("wagtail.embeds.embeds.get_embed")
    def test_direct_call(self, get_embed):
        get_embed.return_value = Embed(html='<img src="http://www.example.com" />')

        result = embed_tag("http://www.youtube.com/watch/")

        self.assertEqual(result, '<img src="http://www.example.com" />')

    @patch("wagtail.embeds.embeds.get_embed")
    def test_call_from_template(self, get_embed):
        get_embed.return_value = Embed(html='<img src="http://www.example.com" />')

        temp = template.Template(
            '{% load wagtailembeds_tags %}{% embed "http://www.youtube.com/watch/" %}'
        )
        result = temp.render(template.Context())

        self.assertEqual(result, '<img src="http://www.example.com" />')

    @patch("wagtail.embeds.embeds.get_embed")
    def test_catches_embed_not_found(self, get_embed):
        get_embed.side_effect = EmbedNotFoundException

        temp = template.Template(
            '{% load wagtailembeds_tags %}{% embed "http://www.youtube.com/watch/" %}'
        )
        result = temp.render(template.Context())

        self.assertEqual(result, "")


class TestEmbedBlock(TestCase):
    def set_up_embed_response(self):
        responses.get(
            url="https://www.youtube.com/oembed",
            match=[
                matchers.query_param_matcher(
                    {"url": "https://www.youtube.com/watch/", "format": "json"}
                ),
            ],
            json={
                "type": "something",
                "url": "http://www.example.com",
                "title": "test_title",
                "width": "640",
                "height": "480",
                "html": "<h1>Hello world!</h1>",
            },
        )

    def test_deserialize(self):
        """
        Deserialising the JSONish value of an EmbedBlock (a URL) should give us an EmbedValue
        for that URL
        """
        block = EmbedBlock(required=False)

        block_val = block.to_python("http://www.example.com/foo")
        self.assertIsInstance(block_val, EmbedValue)
        self.assertEqual(block_val.url, "http://www.example.com/foo")

        # empty values should yield None
        empty_block_val = block.to_python("")
        self.assertIsNone(empty_block_val)

    def test_serialize(self):
        block = EmbedBlock(required=False)

        block_val = EmbedValue("http://www.example.com/foo")
        serialized_val = block.get_prep_value(block_val)
        self.assertEqual(serialized_val, "http://www.example.com/foo")

        serialized_empty_val = block.get_prep_value(None)
        self.assertEqual(serialized_empty_val, "")

    @responses.activate
    def test_render(self):
        self.set_up_embed_response()

        block = EmbedBlock()
        block_val = block.to_python("https://www.youtube.com/watch/")

        temp = template.Template("embed: {{ embed }}")
        context = template.Context({"embed": block_val})
        result = temp.render(context)

        # Check that the embed was in the returned HTML
        self.assertIn("<h1>Hello world!</h1>", result)

    @responses.activate
    def test_render_within_structblock(self):
        """
        When rendering the value of an EmbedBlock directly in a template
        (as happens when accessing it as a child of a StructBlock), the
        proper embed output should be rendered, not the URL.
        """
        self.set_up_embed_response()

        block = blocks.StructBlock(
            [
                ("title", blocks.CharBlock()),
                ("embed", EmbedBlock()),
            ]
        )

        block_val = block.to_python(
            {"title": "A test", "embed": "https://www.youtube.com/watch/"}
        )

        temp = template.Template("embed: {{ self.embed }}")
        context = template.Context({"self": block_val})
        result = temp.render(context)

        self.assertIn("<h1>Hello world!</h1>", result)

    def test_value_from_form(self):
        """
        EmbedBlock should be able to turn a URL submitted as part of a form
        back into an EmbedValue
        """
        block = EmbedBlock(required=False)

        block_val = block.value_from_datadict(
            {"myembed": "http://www.example.com/foo"}, {}, prefix="myembed"
        )
        self.assertIsInstance(block_val, EmbedValue)
        self.assertEqual(block_val.url, "http://www.example.com/foo")

        # empty value should result in None
        empty_val = block.value_from_datadict({"myembed": ""}, {}, prefix="myembed")
        self.assertIsNone(empty_val)

    def test_default(self):
        block1 = EmbedBlock()
        self.assertIsNone(block1.get_default())

        block2 = EmbedBlock(default="")
        self.assertIsNone(block2.get_default())

        block3 = EmbedBlock(default=None)
        self.assertIsNone(block3.get_default())

        block4 = EmbedBlock(default="http://www.example.com/foo")
        self.assertIsInstance(block4.get_default(), EmbedValue)
        self.assertEqual(block4.get_default().url, "http://www.example.com/foo")

        block5 = EmbedBlock(default=EmbedValue("http://www.example.com/foo"))
        self.assertIsInstance(block5.get_default(), EmbedValue)
        self.assertEqual(block5.get_default().url, "http://www.example.com/foo")

        def callable_default():
            return EmbedValue("http://www.example.com/foo")

        block6 = EmbedBlock(default=callable_default)
        self.assertIsInstance(block6.get_default(), EmbedValue)
        self.assertEqual(block6.get_default().url, "http://www.example.com/foo")

    @responses.activate
    def test_clean_required(self):
        self.set_up_embed_response()

        block = EmbedBlock()

        cleaned_value = block.clean(
            block.normalize(EmbedValue("https://www.youtube.com/watch/"))
        )
        self.assertIsInstance(cleaned_value, EmbedValue)
        self.assertEqual(cleaned_value.url, "https://www.youtube.com/watch/")

        with self.assertRaisesMessage(ValidationError, ""):
            block.clean(block.normalize(None))

    @responses.activate
    def test_clean_non_required(self):
        self.set_up_embed_response()

        block = EmbedBlock(required=False)

        cleaned_value = block.clean(
            block.normalize(EmbedValue("https://www.youtube.com/watch/"))
        )
        self.assertIsInstance(cleaned_value, EmbedValue)
        self.assertEqual(cleaned_value.url, "https://www.youtube.com/watch/")

        cleaned_value = block.clean(block.normalize(None))
        self.assertIsNone(cleaned_value)

    @responses.activate
    def test_clean_invalid_url(self):
        # no response set up, so responses will raise a ConnectionError

        non_required_block = EmbedBlock(required=False)

        with self.assertRaises(ValidationError):
            non_required_block.clean(
                non_required_block.normalize(
                    EmbedValue("http://no-oembed-here.com/something")
                )
            )

        required_block = EmbedBlock()

        with self.assertRaises(ValidationError):
            required_block.clean(
                non_required_block.normalize(
                    EmbedValue("http://no-oembed-here.com/something")
                )
            )



# Generated by Django 4.1.7 on 2023-03-16 14:50

from django.db import migrations, models
import django.db.models.deletion
import wagtail.blocks
import wagtail.documents.blocks
import wagtail.fields
import wagtail.images.blocks
import wagtail.snippets.blocks
import wagtail.test.testapp.models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("wagtaildocs", "0012_uploadeddocument"),
        ("wagtailimages", "0025_alter_image_file_alter_rendition_file"),
        ("tests", "0021_remove_non_json_field_streamfields"),
    ]

    operations = [
        migrations.CreateModel(
            name="VariousOnDeleteModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("text", models.TextField()),
                ("object_id", models.UUIDField(blank=True, null=True)),
                (
                    "stream_field",
                    wagtail.fields.StreamField(
                        [
                            (
                                "advertisement_content",
                                wagtail.blocks.StreamBlock(
                                    [
                                        (
                                            "captioned_advert",
                                            wagtail.blocks.StructBlock(
                                                [
                                                    (
                                                        "advert",
                                                        wagtail.snippets.blocks.SnippetChooserBlock(
                                                            wagtail.test.testapp.models.Advert
                                                        ),
                                                    ),
                                                    (
                                                        "caption",
                                                        wagtail.blocks.CharBlock(),
                                                    ),
                                                ]
                                            ),
                                        ),
                                        ("rich_text", wagtail.blocks.RichTextBlock()),
                                    ]
                                ),
                            ),
                            ("image", wagtail.images.blocks.ImageChooserBlock()),
                            (
                                "document",
                                wagtail.documents.blocks.DocumentChooserBlock(),
                            ),
                        ],
                        use_json_field=True,
                    ),
                ),
                ("rich_text", wagtail.fields.RichTextField(blank=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "on_delete_cascade",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="tests.advert",
                    ),
                ),
                (
                    "on_delete_do_nothing",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="tests.advert",
                    ),
                ),
                (
                    "on_delete_protect",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="tests.advert",
                    ),
                ),
                (
                    "on_delete_restrict",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.RESTRICT,
                        related_name="+",
                        to="tests.advert",
                    ),
                ),
                (
                    "on_delete_set",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.SET(
                            wagtail.test.testapp.models.get_default_advert
                        ),
                        related_name="+",
                        to="tests.advert",
                    ),
                ),
                (
                    "on_delete_set_default",
                    models.ForeignKey(
                        blank=True,
                        default=wagtail.test.testapp.models.get_default_advert,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="tests.advert",
                    ),
                ),
                (
                    "on_delete_set_null",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="tests.advert",
                    ),
                ),
                (
                    "protected_document",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="wagtaildocs.document",
                    ),
                ),
                (
                    "protected_image",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="wagtailimages.image",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 2.0.2 on 2018-06-06 18:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailredirects", "0005_capitalizeverbose"),
    ]

    operations = [
        migrations.AlterField(
            model_name="redirect",
            name="redirect_link",
            field=models.URLField(
                blank=True, max_length=255, verbose_name="redirect to any URL"
            ),
        ),
    ]



import uuid

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext

from wagtail.admin.views.generic.preview import PreviewOnEdit as GenericPreviewOnEdit
from wagtail.models import Page


def view_draft(request, page_id):
    page = get_object_or_404(Page, id=page_id).get_latest_revision_as_object()
    perms = page.permissions_for_user(request.user)
    if not (perms.can_publish() or perms.can_edit()):
        raise PermissionDenied

    try:
        preview_mode = request.GET.get("mode", page.default_preview_mode)
    except IndexError:
        raise PermissionDenied

    return page.make_preview_request(request, preview_mode)


class PreviewOnEdit(GenericPreviewOnEdit):
    @property
    def session_key(self):
        return "{}{}".format(self.session_key_prefix, self.kwargs["page_id"])

    def get_object(self):
        page = get_object_or_404(
            Page, id=self.kwargs["page_id"]
        ).get_latest_revision_as_object()
        page_perms = page.permissions_for_user(self.request.user)
        if not page_perms.can_edit():
            raise PermissionDenied
        return page

    def get_form(self, query_dict):
        form_class = self.object.get_edit_handler().get_form_class()
        parent_page = self.object.get_parent().specific

        if not query_dict:
            # Query dict is empty, return null form
            return form_class(
                instance=self.object,
                parent_page=parent_page,
                for_user=self.request.user,
            )

        query_dict = query_dict.copy()

        if not query_dict.get("title"):
            query_dict["title"] = gettext("Placeholder title")

        # Generate a random slug if one is not provided, use UUID to ensure
        # uniqueness without needing to hit the database
        if not query_dict.get("slug"):
            query_dict["slug"] = str(uuid.uuid4())

        return form_class(
            query_dict,
            instance=self.object,
            parent_page=parent_page,
            for_user=self.request.user,
        )


class PreviewOnCreate(PreviewOnEdit):
    @property
    def session_key(self):
        return "{}{}-{}-{}".format(
            self.session_key_prefix,
            self.kwargs["content_type_app_name"],
            self.kwargs["content_type_model_name"],
            self.kwargs["parent_page_id"],
        )

    def get_object(self):
        content_type_app_name = self.kwargs["content_type_app_name"]
        content_type_model_name = self.kwargs["content_type_model_name"]
        parent_page_id = self.kwargs["parent_page_id"]
        try:
            content_type = ContentType.objects.get_by_natural_key(
                content_type_app_name, content_type_model_name
            )
        except ContentType.DoesNotExist:
            raise Http404

        page = content_type.model_class()()
        parent_page = get_object_or_404(Page, id=parent_page_id).specific

        parent_page_perms = parent_page.permissions_for_user(self.request.user)
        if not parent_page_perms.can_add_subpage():
            raise PermissionDenied

        # We need to populate treebeard's path / depth fields in order to
        # pass validation. We can't make these 100% consistent with the rest
        # of the tree without making actual database changes (such as
        # incrementing the parent's numchild field), but by calling treebeard's
        # internal _get_path method, we can set a 'realistic' value that will
        # hopefully enable tree traversal operations
        # to at least partially work.
        page.depth = parent_page.depth + 1
        # Puts the page at the next available path
        # for a child of `parent_page`.
        if parent_page.is_leaf():
            # set the path as the first child of parent_page
            page.path = page._get_path(parent_page.path, page.depth, 1)
        else:
            # add the new page after the last child of parent_page
            page.path = parent_page.get_last_child()._inc_path()

        return page

    def get_form(self, query_dict):
        form = super().get_form(query_dict)
        if self.validate_form(form):
            # Ensures our unsaved page has a suitable url.
            form.instance.set_url_path(form.parent_page)

            form.instance.minimal_clean()
        return form



import hashlib
import unittest

from django.conf import settings
from django.contrib.auth.models import Group, Permission
from django.core import checks
from django.core.cache import caches
from django.core.files import File
from django.core.files.storage import Storage, default_storage, storages
from django.core.files.uploadedfile import SimpleUploadedFile
from django.db.models import Prefetch
from django.db.utils import IntegrityError
from django.test import SimpleTestCase, TestCase, TransactionTestCase, override_settings
from django.urls import reverse
from willow.image import Image as WillowImage

from wagtail.images.exceptions import InvalidFilterSpecError
from wagtail.images.models import (
    Filter,
    Picture,
    Rendition,
    ResponsiveImage,
    SourceImageIOError,
    get_rendition_storage,
)
from wagtail.images.rect import Rect
from wagtail.models import Collection, GroupCollectionPermission, Page, ReferenceIndex
from wagtail.test.testapp.models import (
    CustomRendition,
    EventPage,
    EventPageCarouselItem,
    ReimportedImageModel,
)
from wagtail.test.utils import WagtailTestUtils

from .utils import (
    Image,
    get_test_image_file,
    get_test_image_file_svg,
    get_test_image_filename,
)


class CustomStorage(Storage):
    pass


class TestImage(TestCase):
    def setUp(self):
        # Create an image for running tests on
        self.image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(colour="white"),
        )

    def test_get_image_model_at_import_time(self):
        self.assertEqual(ReimportedImageModel, Image)

    def test_is_portrait(self):
        self.assertFalse(self.image.is_portrait())

    def test_is_landscape(self):
        self.assertTrue(self.image.is_landscape())

    def test_get_rect(self):
        self.assertEqual(self.image.get_rect(), Rect(0, 0, 640, 480))

    def test_get_focal_point(self):
        self.assertIsNone(self.image.get_focal_point())

        # Add a focal point to the image
        self.image.focal_point_x = 100
        self.image.focal_point_y = 200
        self.image.focal_point_width = 50
        self.image.focal_point_height = 20

        # Get it
        self.assertEqual(self.image.get_focal_point(), Rect(75, 190, 125, 210))

    def test_has_focal_point(self):
        self.assertFalse(self.image.has_focal_point())

        # Add a focal point to the image
        self.image.focal_point_x = 100
        self.image.focal_point_y = 200
        self.image.focal_point_width = 50
        self.image.focal_point_height = 20

        self.assertTrue(self.image.has_focal_point())

    def test_set_focal_point(self):
        self.assertIsNone(self.image.focal_point_x)
        self.assertIsNone(self.image.focal_point_y)
        self.assertIsNone(self.image.focal_point_width)
        self.assertIsNone(self.image.focal_point_height)

        self.image.set_focal_point(Rect(100, 150, 200, 350))

        self.assertEqual(self.image.focal_point_x, 150)
        self.assertEqual(self.image.focal_point_y, 250)
        self.assertEqual(self.image.focal_point_width, 100)
        self.assertEqual(self.image.focal_point_height, 200)

        self.image.set_focal_point(None)

        self.assertIsNone(self.image.focal_point_x)
        self.assertIsNone(self.image.focal_point_y)
        self.assertIsNone(self.image.focal_point_width)
        self.assertIsNone(self.image.focal_point_height)

    def test_is_stored_locally(self):
        self.assertTrue(self.image.is_stored_locally())

    @override_settings(
        STORAGES={
            **settings.STORAGES,
            "default": {
                "BACKEND": "wagtail.test.dummy_external_storage.DummyExternalStorage"
            },
        },
    )
    def test_is_stored_locally_with_external_storage(self):
        self.assertFalse(self.image.is_stored_locally())

    def test_get_file_size(self):
        file_size = self.image.get_file_size()
        self.assertIsInstance(file_size, int)
        self.assertGreater(file_size, 0)

    def test_get_file_size_on_missing_file_raises_sourceimageioerror(self):
        self.image.file.delete(save=False)
        with self.assertRaises(SourceImageIOError):
            self.image.get_file_size()

    def test_file_hash(self):
        with self.image.file.open() as f:
            loaded_hash = hashlib.sha1(f.read()).hexdigest()
        # Ensure get_file_hash() (which does the hash by chunks) returns the
        # same hash as the one calculated by loading the file into memory.
        self.assertEqual(self.image.get_file_hash(), loaded_hash)

    def test_get_suggested_focal_point_svg(self):
        """
        Feature detection should not be run on SVGs.

        https://github.com/wagtail/wagtail/issues/11172
        """
        image = Image.objects.create(
            title="Test SVG",
            file=get_test_image_file_svg(),
        )
        self.assertIsNone(image.get_suggested_focal_point())

    def test_default_with_description(self):
        # Primary default should be description
        image = Image.objects.create(
            title="Test Image",
            description="This is a test description",
            file=get_test_image_file(),
        )
        self.assertEqual(image.default_alt_text, image.description)

    def test_default_without_description(self):
        # Secondary default should be title
        image = Image.objects.create(
            title="Test Image",
            file=get_test_image_file(),
        )
        self.assertEqual(image.default_alt_text, image.title)


class TestImageQuerySet(TransactionTestCase):
    fixtures = ["test_empty.json"]

    def test_search_method(self):
        # Create an image for running tests on
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )

        # Search for it
        results = Image.objects.search("Test")
        self.assertEqual(list(results), [image])

    def test_operators(self):
        aaa_image = Image.objects.create(
            title="AAA Test image",
            file=get_test_image_file(),
        )
        zzz_image = Image.objects.create(
            title="ZZZ Test image",
            file=get_test_image_file(),
        )

        results = Image.objects.search("aaa test", operator="and")
        self.assertEqual(list(results), [aaa_image])

        results = Image.objects.search("aaa test", operator="or")
        sorted_results = sorted(results, key=lambda img: img.title)
        self.assertEqual(sorted_results, [aaa_image, zzz_image])

    def test_custom_ordering(self):
        aaa_image = Image.objects.create(
            title="AAA Test image",
            file=get_test_image_file(),
        )
        zzz_image = Image.objects.create(
            title="ZZZ Test image",
            file=get_test_image_file(),
        )

        results = Image.objects.order_by("title").search(
            "Test", order_by_relevance=False
        )
        self.assertEqual(list(results), [aaa_image, zzz_image])
        results = Image.objects.order_by("-title").search(
            "Test", order_by_relevance=False
        )
        self.assertEqual(list(results), [zzz_image, aaa_image])

    def test_search_indexing_prefetches_tags(self):
        for i in range(0, 10):
            image = Image.objects.create(
                title="Test image %d" % i,
                file=get_test_image_file(),
            )
            image.tags.add("aardvark", "artichoke", "armadillo")

        with self.assertNumQueries(2):
            results = {
                image.title: [tag.name for tag in image.tags.all()]
                for image in Image.get_indexed_objects()
            }
            self.assertIn("aardvark", results["Test image 0"])


class TestImagePermissions(WagtailTestUtils, TestCase):
    def setUp(self):
        # Create some user accounts for testing permissions
        self.user = self.create_user(
            username="user", email="user@email.com", password="password"
        )
        self.owner = self.create_user(
            username="owner", email="owner@email.com", password="password"
        )
        self.editor = self.create_user(
            username="editor", email="editor@email.com", password="password"
        )
        self.editor.groups.add(Group.objects.get(name="Editors"))
        self.administrator = self.create_superuser(
            username="administrator",
            email="administrator@email.com",
            password="password",
        )

        # Owner user must have the add_image permission
        image_adders_group = Group.objects.create(name="Image adders")
        GroupCollectionPermission.objects.create(
            group=image_adders_group,
            collection=Collection.get_first_root_node(),
            permission=Permission.objects.get(codename="add_image"),
        )
        self.owner.groups.add(image_adders_group)

        # Create an image for running tests on
        self.image = Image.objects.create(
            title="Test image",
            uploaded_by_user=self.owner,
            file=get_test_image_file(),
        )

    def test_administrator_can_edit(self):
        self.assertTrue(self.image.is_editable_by_user(self.administrator))

    def test_editor_can_edit(self):
        self.assertTrue(self.image.is_editable_by_user(self.editor))

    def test_owner_can_edit(self):
        self.assertTrue(self.image.is_editable_by_user(self.owner))

    def test_user_cant_edit(self):
        self.assertFalse(self.image.is_editable_by_user(self.user))


class TestFilters(SimpleTestCase):
    def test_expand_spec_single(self):
        self.assertEqual(Filter.expand_spec("width-100"), ["width-100"])

    def test_expand_spec_flat(self):
        self.assertEqual(
            Filter.expand_spec("width-100 jpegquality-20"), ["width-100|jpegquality-20"]
        )

    def test_expand_spec_pipe(self):
        self.assertEqual(
            Filter.expand_spec("width-100|jpegquality-20"), ["width-100|jpegquality-20"]
        )

    def test_expand_spec_list(self):
        self.assertEqual(
            Filter.expand_spec(["width-100", "jpegquality-20"]),
            ["width-100|jpegquality-20"],
        )

    def test_expand_spec_braced(self):
        self.assertEqual(
            Filter.expand_spec("width-{100,200}"), ["width-100", "width-200"]
        )

    def test_expand_spec_mixed(self):
        self.assertEqual(
            Filter.expand_spec("width-{100,200} jpegquality-40"),
            ["width-100|jpegquality-40", "width-200|jpegquality-40"],
        )

    def test_expand_spec_mixed_pipe(self):
        self.assertEqual(
            Filter.expand_spec("width-{100,200}|jpegquality-40"),
            ["width-100|jpegquality-40", "width-200|jpegquality-40"],
        )

    def test_expand_spec_multiple_braces(self):
        self.assertEqual(
            Filter.expand_spec("width-{100,200} jpegquality-{40,80} grayscale"),
            [
                "width-100|jpegquality-40|grayscale",
                "width-100|jpegquality-80|grayscale",
                "width-200|jpegquality-40|grayscale",
                "width-200|jpegquality-80|grayscale",
            ],
        )


class TestResponsiveImage(TestCase):
    def setUp(self):
        # Create an image for running tests on
        self.image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        self.rendition_10 = self.image.get_rendition("width-10")

    def test_construct_empty(self):
        img = ResponsiveImage({})
        self.assertEqual(img.renditions, [])
        self.assertEqual(img.attrs, None)

    def test_construct_with_renditions(self):
        renditions = {"a": self.rendition_10}
        img = ResponsiveImage(renditions)
        self.assertEqual(img.renditions, [self.rendition_10])

    def test_evaluate_value(self):
        self.assertFalse(ResponsiveImage({}))
        self.assertFalse(ResponsiveImage({}, {"sizes": "100vw"}))

        renditions = {"a": self.rendition_10}
        self.assertTrue(ResponsiveImage(renditions))

    def test_compare_value(self):
        renditions = {"a": self.rendition_10}
        value1 = ResponsiveImage(renditions)
        value2 = ResponsiveImage(renditions)
        value3 = ResponsiveImage({"a": self.image.get_rendition("width-15")})
        value4 = ResponsiveImage(renditions, {"sizes": "100vw"})
        self.assertNotEqual(value1, value3)
        self.assertNotEqual(value1, 12345)
        self.assertEqual(value1, value2)
        self.assertNotEqual(value1, value4)

    def test_get_width_srcset(self):
        renditions = {
            "width-10": self.rendition_10,
            "width-90": self.image.get_rendition("width-90"),
        }
        filenames = [
            get_test_image_filename(self.image, "width-10"),
            get_test_image_filename(self.image, "width-90"),
        ]
        self.assertEqual(
            ResponsiveImage.get_width_srcset(list(renditions.values())),
            f"{filenames[0]} 10w, {filenames[1]} 90w",
        )

    def test_get_width_srcset_single_rendition(self):
        renditions = {"width-10": self.rendition_10}
        self.assertEqual(
            ResponsiveImage.get_width_srcset(list(renditions.values())),
            get_test_image_filename(self.image, "width-10"),
        )

    def test_render(self):
        renditions = {
            "width-10": self.rendition_10,
            "width-90": self.image.get_rendition("width-90"),
        }
        img = ResponsiveImage(renditions)
        filenames = [
            get_test_image_filename(self.image, "width-10"),
            get_test_image_filename(self.image, "width-90"),
        ]
        self.assertHTMLEqual(
            img.__html__(),
            f"""
                <img
                    alt="Test image"
                    src="{filenames[0]}"
                    srcset="{filenames[0]} 10w, {filenames[1]} 90w"
                    width="10"
                    height="7"
                >
            """,
        )

    def test_render_single_image_same_as_img_tag(self):
        img = ResponsiveImage({"width-10": self.rendition_10})
        self.assertHTMLEqual(img.__html__(), self.rendition_10.img_tag())


class TestPicture(TestCase):
    def setUp(self):
        # Create an image for running tests on
        self.image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        self.rendition_10 = self.image.get_rendition("width-10")

    def test_formats(self):
        renditions = {
            "format-jpeg": self.rendition_10,
            "format-webp": self.rendition_10,
        }
        img = Picture(renditions)
        self.assertEqual(
            img.formats, {"jpeg": [self.rendition_10], "webp": [self.rendition_10]}
        )

    def test_single_format(self):
        renditions = {"format-jpeg": self.rendition_10}
        img = Picture(renditions)
        self.assertEqual(img.formats, {})

    def test_mixed_format(self):
        renditions = {
            "format-jpeg": self.rendition_10,
            "format-webp": self.rendition_10,
            "format-webp-lossless": self.rendition_10,
        }
        img = Picture(renditions)
        self.assertEqual(
            img.formats,
            {
                "jpeg": [self.rendition_10],
                "webp": [self.rendition_10, self.rendition_10],
            },
        )

    def test_fallback_format(self):
        avif = {"format-avif": self.rendition_10}
        webp = {"format-webp": self.rendition_10}
        jpeg = {"format-jpeg": self.rendition_10}
        png = {"format-png": self.rendition_10}
        gif = {"format-gif": self.rendition_10}
        fallbacks = {
            "gif": {**avif, **webp, **jpeg, **png, **gif},
            "png": {**avif, **webp, **jpeg, **png},
            "jpeg": {**avif, **webp, **jpeg},
            "webp": {**avif, **webp},
        }
        for fmt, renditions in fallbacks.items():
            self.assertEqual(Picture(renditions).get_fallback_format(), fmt)

    def test_render_multi_format_sizes(self):
        renditions = {
            "format-jpeg|width-10": self.image.get_rendition("format-jpeg|width-10"),
            "format-jpeg|width-90": self.image.get_rendition("format-jpeg|width-90"),
            "format-webp|width-10": self.image.get_rendition("format-webp|width-10"),
            "format-webp|width-90": self.image.get_rendition("format-webp|width-90"),
        }
        img = Picture(renditions, {"sizes": "100vw"})
        filenames = [
            get_test_image_filename(self.image, "format-jpeg.width-10"),
            get_test_image_filename(self.image, "format-jpeg.width-90"),
            get_test_image_filename(self.image, "format-webp.width-10"),
            get_test_image_filename(self.image, "format-webp.width-90"),
        ]
        self.assertHTMLEqual(
            img.__html__(),
            f"""
                <picture>
                    <source srcset="{filenames[2]} 10w, {filenames[3]} 90w" sizes="100vw" type="image/webp">
                    <img
                        alt="Test image"
                        sizes="100vw"
                        src="{filenames[0]}"
                        srcset="{filenames[0]} 10w, {filenames[1]} 90w"
                        width="10"
                        height="7"
                    >
                </picture>
            """,
        )

    def test_render_single_image_same_as_img_tag(self):
        img = Picture({"width-10": self.rendition_10})
        self.assertHTMLEqual(
            img.__html__(), f"<picture>{self.rendition_10.img_tag()}</picture>"
        )


@override_settings(
    CACHES={"default": {"BACKEND": "django.core.cache.backends.dummy.DummyCache"}}
)
class TestRenditions(TestCase):
    SPECS = ("height-66", "width-100", "width-400")

    def setUp(self):
        # Create an image for running tests on
        self.image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        self.svg_image = Image.objects.create(
            title="Test SVG image",
            file=get_test_image_file_svg(),
        )

    def test_get_rendition_model(self):
        self.assertIs(Image.get_rendition_model(), Rendition)

    def test_stock_rendition_contains_unique_together_constraint(self):
        self.assertEqual([], Rendition.check())

    def test_custom_rendition_may_have_unique_constraint(self):
        self.assertEqual([], CustomRendition.check())

    def test_custom_rendition_without_unique_constraint_raises_error(self):
        custom_rendition_constraints = CustomRendition._meta.constraints
        try:
            CustomRendition._meta.constraints = []
            errors = CustomRendition.check()
        finally:
            CustomRendition._meta.constraints = custom_rendition_constraints

        self.assertEqual(CustomRendition._meta.unique_together, ())
        self.assertEqual(len(errors), 1)
        self.assertIsInstance(errors[0], checks.Error)
        self.assertEqual(errors[0].id, "wagtailimages.E001")
        self.assertEqual(errors[0].obj, CustomRendition)
        self.assertEqual(
            errors[0].msg,
            "Custom rendition model 'tests.CustomRendition' must include a unique "
            "constraint on the 'image', 'filter_spec', and 'focal_point_key' fields.",
        )
        self.assertEqual(
            errors[0].hint,
            "Add models.UniqueConstraint(fields=("
            '"image", "filter_spec", "focal_point_key"), '
            'name="unique_rendition") to CustomRendition.Meta.constraints.',
        )

    def test_minification(self):
        rendition = self.image.get_rendition("width-400")

        # Check size
        self.assertEqual(rendition.width, 400)
        self.assertEqual(rendition.height, 300)

        # check that the rendition has been recorded under the correct filter,
        # via the Rendition.filter_spec attribute (in active use as of Wagtail 1.8)
        self.assertEqual(rendition.filter_spec, "width-400")

    def test_resize_to_max(self):
        rendition = self.image.get_rendition("max-100x100")

        # Check size
        self.assertEqual(rendition.width, 100)
        self.assertEqual(rendition.height, 75)

    def test_resize_to_min(self):
        rendition = self.image.get_rendition("min-120x120")

        # Check size
        self.assertEqual(rendition.width, 160)
        self.assertEqual(rendition.height, 120)

    def test_resize_to_original(self):
        rendition = self.image.get_rendition("original")

        # Check size
        self.assertEqual(rendition.width, 640)
        self.assertEqual(rendition.height, 480)

    def test_cache(self):
        # Get two renditions with the same filter
        first_rendition = self.image.get_rendition("width-400")
        second_rendition = self.image.get_rendition("width-400")

        # Check that they are the same object
        self.assertEqual(first_rendition, second_rendition)

    def test_get_with_filter_instance(self):
        # Get two renditions with the same filter
        first_rendition = self.image.get_rendition("width-400")
        second_rendition = self.image.get_rendition(Filter("width-400"))

        # Check that they are the same object
        self.assertEqual(first_rendition, second_rendition)

    def test_prefetched_rendition_found(self):
        # Request a rendition that does not exist yet
        with self.assertNumQueries(5):
            original_rendition = self.image.get_rendition("width-100")

        # Refetch the same image with all renditions prefetched
        image = Image.objects.prefetch_related("renditions").get(pk=self.image.pk)

        # get_rendition() should return the prefetched rendition
        with self.assertNumQueries(0):
            second_rendition = image.get_rendition("width-100")

        # The renditions should match
        self.assertEqual(original_rendition, second_rendition)

    def test_prefetch_renditions_found(self):
        # Same test as above but uses the `prefetch_renditions` method on the manager instead.
        with self.assertNumQueries(5):
            original_rendition = self.image.get_rendition("width-100")

        image = Image.objects.prefetch_renditions("width-100").get(pk=self.image.pk)

        with self.assertNumQueries(0):
            second_rendition = image.get_rendition("width-100")

        self.assertEqual(original_rendition, second_rendition)

    def test_prefetched_rendition_not_found(self):
        # Request a rendition that does not exist yet
        with self.assertNumQueries(5):
            original_rendition = self.image.get_rendition("width-100")

        # Refetch the same image with all renditions prefetched
        image = Image.objects.prefetch_related("renditions").get(pk=self.image.pk)

        # Request a different rendition from this object
        with self.assertNumQueries(4):
            # The number of queries is fewer than before, because checks for
            # an existing rendition (in cache and db) are skipped
            second_rendition = image.get_rendition("height-66")

        # The renditions should NOT match
        self.assertNotEqual(original_rendition, second_rendition)

        # Request the same rendition again
        with self.assertNumQueries(0):
            # Newly created renditions are appended to the prefetched
            # rendition queryset, so that multiple requests for the same
            # rendition can reuse the same value
            third_rendition = image.get_rendition("height-66")

        # The second and third renditions should be references to the
        # exact same in-memory object
        self.assertIs(second_rendition, third_rendition)

    def test_prefetch_renditions_not_found(self):
        # Same test as above but uses the `prefetch_renditions` method on the manager instead.
        with self.assertNumQueries(5):
            original_rendition = self.image.get_rendition("width-100")

        image = Image.objects.prefetch_renditions("width-100").get(pk=self.image.pk)

        with self.assertNumQueries(4):
            second_rendition = image.get_rendition("height-66")

        self.assertNotEqual(original_rendition, second_rendition)

        with self.assertNumQueries(0):
            third_rendition = image.get_rendition("height-66")

        self.assertIs(second_rendition, third_rendition)

    def test_get_renditions_with_filter_instance(self):
        # Get two renditions with the same filter
        first = list(self.image.get_renditions("width-400").values())
        second = list(self.image.get_renditions(Filter("width-400")).values())

        # Check that they are the same object
        self.assertEqual(first[0], second[0])

    def test_get_renditions_key_order(self):
        # Fetch one of the renditions so it exists before the other two.
        self.image.get_rendition("width-40")
        specs = ["width-30", "width-40", "width-50"]
        renditions_keys = list(self.image.get_renditions(*specs).keys())
        self.assertEqual(renditions_keys, specs)

    def _test_get_renditions_performance(
        self,
        db_queries_expected: int,
        prefetch_restricted: bool = False,
        prefetch_all: bool = False,
    ):
        queryset = Image.objects.all()
        if prefetch_all:
            queryset = queryset.prefetch_related("renditions")
        elif prefetch_restricted:
            queryset = queryset.prefetch_renditions(*self.SPECS)

        image = queryset.get(id=self.image.id)
        with self.assertNumQueries(db_queries_expected):
            image.get_renditions(*self.SPECS)

    def test_get_renditions_performance_with_rendition_caching_disabled(self):
        # ATTEMPT 1
        # 1) An initial lookup for rendition from the DB
        # 2) A check for clashes before bulk saving new renditions
        # 3) A bulk_create() to save new renditions
        self._test_get_renditions_performance(3)

        # ATTEMPT 2
        # With all renditions already created, we should just see
        # 1) An initial lookup for rendition from the DB
        self._test_get_renditions_performance(1)

        # ATTEMPT 3
        # If the existing renditions are prefetched, no further queries should
        # be needed, whether that's with prefetch_related("renditions") or
        # prefetch_renditions()
        self._test_get_renditions_performance(0, prefetch_all=True)
        self._test_get_renditions_performance(0, prefetch_restricted=True)

    @override_settings(
        CACHES={
            "renditions": {
                "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
            },
        },
    )
    def test_get_renditions_performance_with_rendition_caching_enabled(self):
        # ATTEMPT 1
        # 1) An initial lookup for rendition from the DB
        # 2) A check for clashes before bulk saving new renditions
        # 3) A bulk_create() to save new renditions
        self._test_get_renditions_performance(3)

        # ATTEMPT 2
        # Any renditions created in the first attempt should have
        # been added to the cache, so a second request should bypass
        # the database completely
        self._test_get_renditions_performance(0)

        # ATTEMPT 3
        # Prefetching renditions should mean that no queries are
        # required, and no cache hits are made
        self._test_get_renditions_performance(0, prefetch_all=True)

    def test_create_renditions(self):
        filter_list = [Filter(spec) for spec in self.SPECS]
        # When no renditions exist, there should be one query for
        # 'clash identification', and another for bulk creation
        with self.assertNumQueries(2):
            result = self.image.create_renditions(*filter_list)

        # Renditions should match the filters
        for filter, rendition in result.items():
            self.assertEqual(filter.spec, rendition.filter_spec)

        # Filter specs should match the filters that were provided as arguments
        self.assertEqual(
            {filter.spec for filter in result.keys()},
            {filter.spec for filter in filter_list},
        )

        # When the renditions already exist, there should be one query
        # for 'clash identification', but that is all
        with self.assertNumQueries(1):
            result = self.image.create_renditions(*filter_list)

        # Renditions should match the filters
        for filter, rendition in result.items():
            self.assertEqual(filter.spec, rendition.filter_spec)

        # Filter specs should match the filters that were provided as arguments
        self.assertEqual(
            {filter.spec for filter in result.keys()},
            {filter.spec for filter in filter_list},
        )

        # Another request should give us an equal result
        with self.assertNumQueries(1):
            second_result = self.image.create_renditions(*filter_list)
        self.assertEqual(result, second_result)

        # When only some renditions exist, we should see the create query
        # once again
        self.image.renditions.filter(filter_spec=self.SPECS[0]).delete()
        with self.assertNumQueries(2):
            third_result = self.image.create_renditions(*filter_list)

        # Equality check should fail, as newly created rendition has a different pk
        self.assertNotEqual(third_result, result)

        # But, we should see equality on the keys
        self.assertEqual(third_result.keys(), result.keys())

    def test_alt_attribute(self):
        rendition = self.image.get_rendition("width-400")
        self.assertEqual(rendition.alt, "Test image")

    def test_full_url(self):
        ren_img = self.image.get_rendition("original")
        full_url = ren_img.full_url
        img_name = ren_img.file.name.split("/")[1]
        self.assertEqual(full_url, f"http://testserver/media/images/{img_name}")

    @override_settings(
        CACHES={
            "renditions": {
                "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
            },
        },
    )
    def test_renditions_cache(self):
        cache = Rendition.cache_backend
        rendition = self.image.get_rendition("width-500")
        rendition_cache_key = rendition.get_cache_key()

        # Check rendition is saved to cache
        self.assertEqual(cache.get(rendition_cache_key), rendition)

        # Mark a rendition to check it comes from cache
        rendition._mark = "original"
        cache.set(rendition_cache_key, rendition)

        # Check if get_rendition returns the rendition from cache
        with self.assertNumQueries(0):
            new_rendition = self.image.get_rendition("width-500")
        self.assertEqual(new_rendition._mark, "original")

        # But, not if the rendition has been prefetched
        fresh_image = Image.objects.prefetch_related("renditions").get(pk=self.image.pk)
        with self.assertNumQueries(0):
            prefetched_rendition = fresh_image.get_rendition("width-500")
        self.assertFalse(hasattr(prefetched_rendition, "_mark"))

        # Check that the image instance is the same as the retrieved rendition
        self.assertIs(new_rendition.image, self.image)

        # changing the image file should invalidate the cache
        self.image.file = get_test_image_file(colour="green")
        self.image.save()
        # deleting renditions would normally happen within the 'edit' view on file change -
        # we're bypassing that here, so have to do it manually
        self.image.renditions.all().delete()
        new_rendition = self.image.get_rendition("width-500")
        self.assertFalse(hasattr(new_rendition, "_mark"))

        # changing it back should also generate a new rendition and not re-use
        # the original one (because that file has now been deleted in the change)
        self.image.file = get_test_image_file(colour="white")
        self.image.save()
        self.image.renditions.all().delete()
        new_rendition = self.image.get_rendition("width-500")
        self.assertFalse(hasattr(new_rendition, "_mark"))

    def test_prefers_rendition_cache_backend(self):
        with override_settings(
            CACHES={
                "default": {
                    "BACKEND": "django.core.cache.backends.dummy.DummyCache",
                },
                "renditions": {
                    "BACKEND": "django.core.cache.backends.dummy.DummyCache",
                },
            }
        ):
            self.assertEqual(Rendition.cache_backend, caches["renditions"])

    def test_uses_default_cache_when_no_renditions_cache(self):
        with override_settings(
            CACHES={
                "default": {
                    "BACKEND": "django.core.cache.backends.dummy.DummyCache",
                }
            }
        ):
            self.assertEqual(Rendition.cache_backend, caches["default"])

    def test_focal_point(self):
        self.image.focal_point_x = 100
        self.image.focal_point_y = 200
        self.image.focal_point_width = 50
        self.image.focal_point_height = 20
        self.image.save()

        # Generate a rendition that's half the size of the original
        rendition = self.image.get_rendition("width-320")

        self.assertEqual(rendition.focal_point.round(), Rect(37, 95, 63, 105))
        self.assertEqual(rendition.focal_point.centroid.x, 50)
        self.assertEqual(rendition.focal_point.centroid.y, 100)
        self.assertEqual(rendition.focal_point.width, 25)
        self.assertEqual(rendition.focal_point.height, 10)

        self.assertEqual(
            rendition.background_position_style, "background-position: 15% 41%;"
        )

        # Individual background position properties
        self.assertEqual(rendition.background_position_x, "15%")
        self.assertEqual(rendition.background_position_y, "41%")

    def test_background_position_style_default(self):
        # Generate a rendition that's half the size of the original
        rendition = self.image.get_rendition("width-320")

        self.assertEqual(
            rendition.background_position_style, "background-position: 50% 50%;"
        )

        # Individual background position properties
        self.assertEqual(rendition.background_position_x, "50%")
        self.assertEqual(rendition.background_position_y, "50%")

    @override_settings()
    def test_rendition_storage_setting_absent(self):
        del settings.WAGTAILIMAGES_RENDITION_STORAGE
        self.assertFalse(hasattr(settings, "WAGTAILIMAGES_RENDITION_STORAGE"))
        self.assertEqual(get_rendition_storage(), default_storage)

    @override_settings(
        WAGTAILIMAGES_RENDITION_STORAGE="wagtail.images.tests.test_models.CustomStorage"
    )
    def test_rendition_storage_setting_given_dotted_path(self):
        self.assertIsInstance(get_rendition_storage(), CustomStorage)

    @override_settings(WAGTAILIMAGES_RENDITION_STORAGE=CustomStorage())
    def test_rendition_storage_setting_given_storage_instance(self):
        self.assertEqual(
            get_rendition_storage(), settings.WAGTAILIMAGES_RENDITION_STORAGE
        )

    @override_settings(
        STORAGES={
            "default": {
                "BACKEND": "django.core.files.storage.FileSystemStorage",
            },
            "staticfiles": {
                "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
            },
            "custom_storage": {
                "BACKEND": "wagtail.images.tests.test_models.CustomStorage",
            },
        },
        WAGTAILIMAGES_RENDITION_STORAGE="custom_storage",
    )
    def test_rendition_storage_setting_given_storage_alias(self):
        self.assertEqual(
            get_rendition_storage(), storages[settings.WAGTAILIMAGES_RENDITION_STORAGE]
        )

    def test_image_get_rendition_preserve_svg(self):
        image_rendition_1 = self.image.get_rendition(
            "width-400|bgcolor-000|format-jpeg|preserve-svg"
        )
        # no directives stripped except 'preserve-svg'
        self.assertEqual(
            image_rendition_1.filter_spec, "width-400|bgcolor-000|format-jpeg"
        )

        image_rendition_2 = self.image.get_rendition(
            "width-400|bgcolor-000|format-jpeg"
        )
        # preserve-svg has no effect and thus the existing rendition will be returned
        self.assertEqual(image_rendition_1, image_rendition_2)

        # same behaviour when passing a Filter object rather than a string
        image_rendition_3 = self.image.get_rendition(
            Filter("width-400|bgcolor-000|format-jpeg|preserve-svg")
        )
        self.assertEqual(image_rendition_1, image_rendition_3)
        image_rendition_4 = self.image.get_rendition(
            Filter("width-400|bgcolor-000|format-jpeg")
        )
        self.assertEqual(image_rendition_1, image_rendition_4)

        # preserve-svg with no other directive raises error
        with self.assertRaises(InvalidFilterSpecError):
            self.image.get_rendition("preserve-svg")

    def test_svg_get_rendition_preserve_svg(self):
        # rasterize directives stripped
        svg_rendition_1 = self.svg_image.get_rendition(
            "width-400|bgcolor-000|format-jpeg|preserve-svg"
        )
        self.assertEqual(svg_rendition_1.filter_spec, "width-400")

        # str filter & Filter with preserve-svg produce same result
        svg_rendition_2 = self.svg_image.get_rendition(
            Filter("width-400|bgcolor-000|format-jpeg|preserve-svg")
        )
        self.assertEqual(svg_rendition_1, svg_rendition_2)

        # fallback to 'original' if no SVG-safe directives remain
        svg_rendition_3 = self.svg_image.get_rendition(
            Filter("bgcolor-000|format-jpeg|preserve-svg")
        )
        self.assertEqual(svg_rendition_3.filter_spec, "original")

        # no rasterize directives, no preserve-svg
        svg_rendition_4 = self.svg_image.get_rendition(Filter("width-400"))
        self.assertEqual(svg_rendition_4.filter_spec, "width-400")

        # no rasterize directives, preserve-svg has no affect
        svg_rendition_5 = self.svg_image.get_rendition(Filter("width-400|preserve-svg"))
        self.assertEqual(svg_rendition_4, svg_rendition_5)

        # has raterize directives but no preserve-svg raises error
        with self.assertRaises(AttributeError):
            self.svg_image.get_rendition(Filter("width-400|bgcolor-000|format-jpeg"))

    def test_image_get_renditions_preserve_svg(self):
        renditions = self.image.get_renditions(
            "width-400|bgcolor-000|format-jpeg|preserve-svg",
            "width-200|bgcolor-000|format-jpeg|preserve-svg",
        )
        filename1 = get_test_image_filename(
            self.image, "width-400.bgcolor-000.format-jpeg"
        )
        filename2 = get_test_image_filename(
            self.image, "width-200.bgcolor-000.format-jpeg"
        )

        # no directives stripped except 'preserve-svg'
        # (which is stripped from both the dictionary key and the resulting rendition)
        self.assertEqual(
            renditions["width-400|bgcolor-000|format-jpeg"].filter_spec,
            "width-400|bgcolor-000|format-jpeg",
        )
        self.assertEqual(renditions["width-400|bgcolor-000|format-jpeg"].url, filename1)

        self.assertEqual(
            renditions["width-200|bgcolor-000|format-jpeg"].filter_spec,
            "width-200|bgcolor-000|format-jpeg",
        )
        self.assertEqual(renditions["width-200|bgcolor-000|format-jpeg"].url, filename2)

    def test_svg_get_renditions_preserve_svg(self):
        renditions = self.svg_image.get_renditions(
            "width-400|bgcolor-000|format-jpeg|preserve-svg",
            "width-200|bgcolor-000|format-jpeg|preserve-svg",
        )
        filename1 = get_test_image_filename(self.svg_image, "width-400")
        filename2 = get_test_image_filename(self.svg_image, "width-200")

        # all non-SVG-safe directives stripped (from both the dictionary key and the resulting rendition)
        self.assertEqual(renditions["width-400"].filter_spec, "width-400")
        self.assertEqual(renditions["width-400"].url, filename1)

        self.assertEqual(renditions["width-200"].filter_spec, "width-200")
        self.assertEqual(renditions["width-200"].url, filename2)


@override_settings(
    CACHES={"default": {"BACKEND": "django.core.cache.backends.dummy.DummyCache"}}
)
class TestPrefetchRenditions(TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        self.images = []
        self.event_pages_pks = []

        event_pages = EventPage.objects.all()[:3]
        for i, page in enumerate(event_pages):
            page.feed_image = image = Image.objects.create(
                title="Test image {i}",
                file=get_test_image_file(),
            )
            page.save(update_fields=["feed_image"])
            self.images.append(image)
            self.event_pages_pks.append(page.pk)

        # Generate renditions
        self.small_renditions = [
            image.get_rendition("max-100x100") for image in self.images
        ]
        self.large_renditions = [
            image.get_rendition("min-300x600") for image in self.images
        ]

    def test_prefetch_renditions_on_non_image_querysets(self):
        prefetch_images_and_small_renditions = Prefetch(
            "feed_image", queryset=Image.objects.prefetch_renditions("max-100x100")
        )
        with self.assertNumQueries(3):
            # One query to get the `EventPage`s, another one to fetch the feed images
            # and a last one to select matching renditions.
            pages = list(
                EventPage.objects.prefetch_related(
                    prefetch_images_and_small_renditions
                ).filter(pk__in=self.event_pages_pks)
            )

        with self.assertNumQueries(0):
            # No additional query since small renditions were prefetched.
            small_renditions = [
                page.feed_image.get_rendition("max-100x100") for page in pages
            ]

        self.assertListEqual(self.small_renditions, small_renditions)

        with self.assertNumQueries(3):
            # Additional queries since large renditions weren't prefetched.
            large_renditions = [
                page.feed_image.get_rendition("min-300x600") for page in pages
            ]

        self.assertListEqual(self.large_renditions, large_renditions)


class TestUsageCount(TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        self.image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )

    def test_unused_image_usage_count(self):
        self.assertEqual(self.image.get_usage().count(), 0)

    def test_used_image_document_usage_count(self):
        with self.captureOnCommitCallbacks(execute=True):
            page = EventPage.objects.get(id=4)
            event_page_carousel_item = EventPageCarouselItem()
            event_page_carousel_item.page = page
            event_page_carousel_item.image = self.image
            event_page_carousel_item.save()
        self.assertEqual(self.image.get_usage().count(), 1)


class TestGetUsage(TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        self.image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )

    def test_unused_image_get_usage(self):
        self.assertEqual(list(self.image.get_usage()), [])

    def test_used_image_document_get_usage(self):
        with self.captureOnCommitCallbacks(execute=True):
            page = EventPage.objects.get(id=4)
            event_page_carousel_item = EventPageCarouselItem()
            event_page_carousel_item.page = page
            event_page_carousel_item.image = self.image
            event_page_carousel_item.save()

        self.assertIsInstance(self.image.get_usage()[0], tuple)
        self.assertIsInstance(self.image.get_usage()[0][0], Page)
        self.assertIsInstance(self.image.get_usage()[0][1], list)
        self.assertIsInstance(self.image.get_usage()[0][1][0], ReferenceIndex)


class TestGetWillowImage(TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        self.image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )

    def test_willow_image_object_returned(self):
        with self.image.get_willow_image() as willow_image:
            self.assertIsInstance(willow_image, WillowImage)

    def test_with_missing_image(self):
        # Image id=1 in test fixtures has a missing image file
        bad_image = Image.objects.get(id=1)

        # Attempting to get the Willow image for images without files
        # should raise a SourceImageIOError
        with self.assertRaises(SourceImageIOError):
            with bad_image.get_willow_image():
                self.fail()  # Shouldn't get here

    def test_closes_image(self):
        # This tests that willow closes images after use
        with self.image.get_willow_image():
            self.assertFalse(self.image.file.closed)

        self.assertTrue(self.image.file.closed)

    def test_closes_image_on_exception(self):
        # This tests that willow closes images when the with is exited with an exception
        try:
            with self.image.get_willow_image():
                self.assertFalse(self.image.file.closed)
                raise ValueError("Something went wrong!")
        except ValueError:
            pass

        self.assertTrue(self.image.file.closed)

    def test_doesnt_close_open_image(self):
        # This tests that when the image file is already open, get_willow_image doesn't close it (#1256)
        self.image.file.open("rb")

        with self.image.get_willow_image():
            pass

        self.assertFalse(self.image.file.closed)

        self.image.file.close()


class TestIssue573(TestCase):
    """
    This tests for a bug which causes filename limit on Renditions to be reached
    when the Image has a long original filename and a big focal point key
    """

    def test_issue_573(self):
        # Create an image with a big filename and focal point
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(
                "thisisaverylongfilename-cdefghijklmnopqrstuvwxyzab-supercalifragilisticexpialidocious.png"
            ),
            focal_point_x=1000,
            focal_point_y=1000,
            focal_point_width=1000,
            focal_point_height=1000,
        )

        # Try creating a rendition from that image
        # This would crash if the bug is present
        image.get_rendition("fill-800x600")


@override_settings(_WAGTAILSEARCH_FORCE_AUTO_UPDATE=["elasticsearch"])
class TestIssue613(WagtailTestUtils, TestCase):
    def get_elasticsearch_backend(self):
        from django.conf import settings

        from wagtail.search.backends import get_search_backend

        if "elasticsearch" not in settings.WAGTAILSEARCH_BACKENDS:
            raise unittest.SkipTest("No elasticsearch backend active")

        return get_search_backend("elasticsearch")

    def setUp(self):
        self.search_backend = self.get_elasticsearch_backend()
        self.login()

    def add_image(self, **params):
        post_data = {
            "title": "Test image",
            "file": SimpleUploadedFile(
                "test.png", get_test_image_file().file.getvalue()
            ),
        }
        post_data.update(params)
        response = self.client.post(reverse("wagtailimages:add"), post_data)

        # Should redirect back to index
        self.assertRedirects(response, reverse("wagtailimages:index"))

        # Check that the image was created
        images = Image.objects.filter(title="Test image")
        self.assertEqual(images.count(), 1)

        # Test that size was populated correctly
        image = images.first()
        self.assertEqual(image.width, 640)
        self.assertEqual(image.height, 480)

        return image

    def edit_image(self, **params):
        # Create an image to edit
        self.image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )

        # Edit it
        post_data = {
            "title": "Edited",
        }
        post_data.update(params)
        response = self.client.post(
            reverse("wagtailimages:edit", args=(self.image.id,)), post_data
        )

        # Should redirect back to index
        self.assertRedirects(response, reverse("wagtailimages:index"))

        # Check that the image was edited
        image = Image.objects.get(id=self.image.id)
        self.assertEqual(image.title, "Edited")
        return image

    def test_issue_613_on_add(self):
        # Reset the search index
        self.search_backend.reset_index()
        self.search_backend.add_type(Image)

        # Add an image with some tags
        image = self.add_image(tags="hello")
        self.search_backend.refresh_index()

        # Search for it by tag
        results = self.search_backend.search("hello", Image)

        # Check
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, image.id)

    def test_issue_613_on_edit(self):
        # Reset the search index
        self.search_backend.reset_index()
        self.search_backend.add_type(Image)

        # Add an image with some tags
        image = self.edit_image(tags="hello")
        self.search_backend.refresh_index()

        # Search for it by tag
        results = self.search_backend.search("hello", Image)

        # Check
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, image.id)


class TestIssue312(TestCase):
    def test_duplicate_renditions(self):
        # Create an image
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )

        # Get two renditions and check that they're the same
        rend1 = image.get_rendition("fill-100x100")
        rend2 = image.get_rendition("fill-100x100")
        self.assertEqual(rend1, rend2)

        # Now manually duplicate the renditon and check that the database blocks it
        self.assertRaises(
            IntegrityError,
            Rendition.objects.create,
            image=rend1.image,
            filter_spec=rend1.filter_spec,
            width=rend1.width,
            height=rend1.height,
            focal_point_key=rend1.focal_point_key,
        )


class TestFilenameReduction(TestCase):
    """
    This tests for a bug which results in filenames without extensions
    causing an infinite loop
    """

    def test_filename_reduction_no_ext(self):
        # Create an image with a big filename and no extension
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(
                "thisisaverylongfilename-defghijklmnopqrstuvwxyzabc-supercalifragilisticexpialidocioussuperlong"
            ),
        )

        # Saving file will result in infinite loop when bug is present
        image.save()
        self.assertEqual(
            "original_images/thisisaverylongfilename-defghijklmnopqrstuvwxyzabc-supercalifragilisticexpiali",
            image.file.name,
        )

    # Test for happy path. Long filename with extension
    def test_filename_reduction_ext(self):
        # Create an image with a big filename and extensions
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(
                "thisisaverylongfilename-efghijklmnopqrstuvwxyzabcd-supercalifragilisticexpialidocioussuperlong.png"
            ),
        )

        image.save()
        self.assertEqual(
            "original_images/thisisaverylongfilename-efghijklmnopqrstuvwxyzabcd-supercalifragilisticexp.png",
            image.file.name,
        )


class TestRenditionOrientation(TestCase):
    """
    This tests for a bug where images with exif orientations which
    required rotation for display were cropped and sized based on the
    unrotated image dimensions.

    For example images with specified dimensions of 640x450 but an exif orientation of 6
    should appear as a 450x640 portrait, but instead were still cropped to 640x450.

    Actual image files are used so that exif orientation data will exist for the rotation to function correctly.
    """

    def assert_orientation_landscape_image_is_correct(self, rendition):
        """
        Check that the image has the correct colored pixels in the right places
        so that we know the image did not physically rotate.
        """

        from willow.plugins.pillow import PillowImage

        with rendition.get_willow_image() as willow_image:
            image = PillowImage.open(willow_image)
        # Check that the image is the correct size (and not rotated)
        self.assertEqual(image.get_size(), (600, 450))
        # Check that the red flower is in the bottom left
        # The JPEGs have compressed slightly differently so the colours won't be spot on
        colour = image.image.convert("RGB").getpixel((155, 282))
        self.assertAlmostEqual(colour[0], 217, delta=25)
        self.assertAlmostEqual(colour[1], 38, delta=25)
        self.assertAlmostEqual(colour[2], 46, delta=25)

        # Check that the water is at the bottom
        colour = image.image.convert("RGB").getpixel((377, 434))
        self.assertAlmostEqual(colour[0], 85, delta=25)
        self.assertAlmostEqual(colour[1], 93, delta=25)
        self.assertAlmostEqual(colour[2], 65, delta=25)

    def test_jpeg_with_orientation_1(self):
        with open("wagtail/images/tests/image_files/landscape_1.jpg", "rb") as f:
            image = Image.objects.create(title="Test image", file=File(f))

        # check preconditions
        self.assertEqual(image.width, 600)
        self.assertEqual(image.height, 450)
        rendition = image.get_rendition("original")
        # Check dimensions stored on the model
        self.assertEqual(rendition.width, 600)
        self.assertEqual(rendition.height, 450)
        # Check actual image dimensions and orientation
        self.assert_orientation_landscape_image_is_correct(rendition)

    def test_jpeg_with_orientation_2(self):
        with open("wagtail/images/tests/image_files/landscape_2.jpg", "rb") as f:
            image = Image.objects.create(title="Test image", file=File(f))

        # check preconditions
        self.assertEqual(image.width, 600)
        self.assertEqual(image.height, 450)
        rendition = image.get_rendition("original")
        # Check dimensions stored on the model
        self.assertEqual(rendition.width, 600)
        self.assertEqual(rendition.height, 450)
        # Check actual image dimensions and orientation
        self.assert_orientation_landscape_image_is_correct(rendition)

    def test_jpeg_with_orientation_3(self):
        with open("wagtail/images/tests/image_files/landscape_3.jpg", "rb") as f:
            image = Image.objects.create(title="Test image", file=File(f))

        # check preconditions
        self.assertEqual(image.width, 600)
        self.assertEqual(image.height, 450)
        rendition = image.get_rendition("original")
        # Check dimensions stored on the model
        self.assertEqual(rendition.width, 600)
        self.assertEqual(rendition.height, 450)
        # Check actual image dimensions and orientation
        self.assert_orientation_landscape_image_is_correct(rendition)

    def test_jpeg_with_orientation_4(self):
        with open("wagtail/images/tests/image_files/landscape_4.jpg", "rb") as f:
            image = Image.objects.create(title="Test image", file=File(f))

        # check preconditions
        self.assertEqual(image.width, 600)
        self.assertEqual(image.height, 450)
        rendition = image.get_rendition("original")
        # Check dimensions stored on the model
        self.assertEqual(rendition.width, 600)
        self.assertEqual(rendition.height, 450)
        # Check actual image dimensions and orientation
        self.assert_orientation_landscape_image_is_correct(rendition)

    # tests below here have a specified width x height in portrait but
    # an orientation specified of landscape, so the original shows a height > width
    # but the rendition is corrected to height < width.
    def test_jpeg_with_orientation_5(self):
        with open("wagtail/images/tests/image_files/landscape_6.jpg", "rb") as f:
            image = Image.objects.create(title="Test image", file=File(f))

        # check preconditions
        self.assertEqual(image.width, 450)
        self.assertEqual(image.height, 600)
        rendition = image.get_rendition("original")
        # Check dimensions stored on the model
        self.assertEqual(rendition.width, 600)
        self.assertEqual(rendition.height, 450)
        # Check actual image dimensions and orientation
        self.assert_orientation_landscape_image_is_correct(rendition)

    def test_jpeg_with_orientation_6(self):
        with open("wagtail/images/tests/image_files/landscape_6.jpg", "rb") as f:
            image = Image.objects.create(title="Test image", file=File(f))

        # check preconditions
        self.assertEqual(image.width, 450)
        self.assertEqual(image.height, 600)
        rendition = image.get_rendition("original")
        # Check dimensions stored on the model
        self.assertEqual(rendition.width, 600)
        self.assertEqual(rendition.height, 450)
        # Check actual image dimensions and orientation
        self.assert_orientation_landscape_image_is_correct(rendition)

    def test_jpeg_with_orientation_7(self):
        with open("wagtail/images/tests/image_files/landscape_7.jpg", "rb") as f:
            image = Image.objects.create(title="Test image", file=File(f))

        # check preconditions
        self.assertEqual(image.width, 450)
        self.assertEqual(image.height, 600)
        rendition = image.get_rendition("original")
        # Check dimensions stored on the model
        self.assertEqual(rendition.width, 600)
        self.assertEqual(rendition.height, 450)
        # Check actual image dimensions and orientation
        self.assert_orientation_landscape_image_is_correct(rendition)

    def test_jpeg_with_orientation_8(self):
        with open("wagtail/images/tests/image_files/landscape_8.jpg", "rb") as f:
            image = Image.objects.create(title="Test image", file=File(f))

        # check preconditions
        self.assertEqual(image.width, 450)
        self.assertEqual(image.height, 600)
        rendition = image.get_rendition("original")
        # Check dimensions stored on the model
        self.assertEqual(rendition.width, 600)
        self.assertEqual(rendition.height, 450)
        # Check actual image dimensions and orientation
        self.assert_orientation_landscape_image_is_correct(rendition)



from collections import OrderedDict

from rest_framework.fields import Field

from ..models import SourceImageIOError
from ..utils import to_svg_safe_spec


class ImageRenditionField(Field):
    """
    A field that generates a rendition with the specified filter spec, and serialises
    details of that rendition.

    Example:
    "thumbnail": {
        "url": "/media/images/myimage.max-165x165.jpg",
        "full_url": "https://media.example.com/media/images/myimage.max-165x165.jpg",
        "width": 165,
        "height": 100,
        "alt": "Image alt text"
    }

    If there is an error with the source image. The dict will only contain a single
    key, "error", indicating this error:

    "thumbnail": {
        "error": "SourceImageIOError"
    }
    """

    def __init__(self, filter_spec, preserve_svg=False, *args, **kwargs):
        self.filter_spec = filter_spec
        self.preserve_svg = preserve_svg
        super().__init__(*args, **kwargs)

    def to_representation(self, image):
        try:
            if image.is_svg() and self.preserve_svg:
                filter_spec = to_svg_safe_spec(self.filter_spec)
            else:
                filter_spec = self.filter_spec

            thumbnail = image.get_rendition(filter_spec)

            return OrderedDict(
                [
                    ("url", thumbnail.url),
                    ("full_url", thumbnail.full_url),
                    ("width", thumbnail.width),
                    ("height", thumbnail.height),
                    ("alt", thumbnail.alt),
                ]
            )
        except SourceImageIOError:
            return OrderedDict(
                [
                    ("error", "SourceImageIOError"),
                ]
            )



from django.core.exceptions import PermissionDenied

from wagtail.models import ReferenceIndex


class ReferenceIndexMixin:
    def annotate_items(self, items):
        items = super().annotate_items(items)
        self.object_references = ReferenceIndex.get_grouped_references_to_in_bulk(items)
        self.is_protected = False
        for item in self.object_references:
            item._usage_references = self.object_references.get(item)
            self.is_protected = self.is_protected or item._usage_references.is_protected
        return items

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["is_protected"] = self.is_protected
        return context

    def get_usage_url(self, item):
        return None

    def object_context(self, item):
        context = super().object_context(item)
        context.update(
            {
                "usage_count": (item._usage_references.count()),
                "usage_url": self.get_usage_url(item),
                "is_protected": item._usage_references.is_protected,
            }
        )
        return context

    def prepare_action(self, objects, objects_without_access):
        super().prepare_action(objects, objects_without_access)
        if self.is_protected:
            raise PermissionDenied



# Generated by Django 5.1.5 on 2025-02-28 00:28

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("tests", "0047_advertwithcustomuuidprimarykey_page"),
        ("wagtailcore", "0094_alter_page_locale"),
    ]

    operations = [
        migrations.CreateModel(
            name="RequiredDatePage",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="wagtailcore.page",
                    ),
                ),
                ("deadline", models.DateField()),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page",),
        ),
    ]



"""
Draftail / contentstate conversion
"""

from draftjs_exporter.dom import DOM

from wagtail.admin.rich_text.converters.contentstate_models import Entity
from wagtail.admin.rich_text.converters.html_to_contentstate import (
    AtomicBlockEntityElementHandler,
)
from wagtail.embeds import embeds
from wagtail.embeds.exceptions import EmbedException


def media_embed_entity(props):
    """
    Helper to construct elements of the form
    <embed embedtype="media" url="https://www.youtube.com/watch?v=y8Kyi0WNg40"/>
    when converting from contentstate data
    """
    return DOM.create_element(
        "embed",
        {
            "embedtype": "media",
            "url": props.get("url"),
        },
    )


class MediaEmbedElementHandler(AtomicBlockEntityElementHandler):
    """
    Rule for building an embed entity when converting from database representation
    to contentstate
    """

    def create_entity(self, name, attrs, state, contentstate):
        try:
            embed_obj = embeds.get_embed(attrs["url"])
            embed_data = {
                "embedType": embed_obj.type,
                "url": embed_obj.url,
                "providerName": embed_obj.provider_name,
                "authorName": embed_obj.author_name,
                "thumbnail": embed_obj.thumbnail_url,
                "title": embed_obj.title,
            }
        except EmbedException:
            embed_data = {"url": attrs["url"]}
        return Entity("EMBED", "IMMUTABLE", embed_data)


ContentstateMediaConversionRule = {
    "from_database_format": {
        'embed[embedtype="media"]': MediaEmbedElementHandler(),
    },
    "to_database_format": {"entity_decorators": {"EMBED": media_embed_entity}},
}



from django.contrib.contenttypes.models import ContentType


def get_default_page_content_type():
    """
    Returns the content type to use as a default for pages whose content type
    has been deleted.
    """
    from wagtail.models import Page

    return ContentType.objects.get_for_model(Page)



from django import template
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag(takes_context=True)
def cell_classname(context, row_index, col_index, table_header=None):
    classnames = context.get("classnames")
    if classnames:
        if table_header is not None:
            row_index += 1
        index = (row_index, col_index)
        cell_class = classnames.get(index)
        if cell_class:
            return mark_safe(f'class="{cell_class}"')
    return ""


@register.simple_tag(takes_context=True)
def cell_hidden(context, row_index, col_index, table_header=None):
    hidden = context.get("hidden")
    if hidden:
        if table_header is not None:
            row_index += 1
        index = (row_index, col_index)
        return hidden.get(index, False)
    return False


@register.simple_tag(takes_context=True)
def cell_span(context, row_index, col_index, table_header=None):
    spans = context.get("spans")
    if spans:
        if table_header is not None:
            row_index += 1
        index = (row_index, col_index)
        cell_span = spans.get(index)
        if cell_span:
            return mark_safe(
                'rowspan="{}" colspan="{}"'.format(
                    cell_span["rowspan"],
                    cell_span["colspan"],
                )
            )
    return ""



import logging

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

logger = logging.getLogger("wagtail.frontendcache")


class InvalidFrontendCacheBackendError(ImproperlyConfigured):
    pass


def get_backends(backend_settings=None, backends=None):
    # Get backend settings from WAGTAILFRONTENDCACHE setting
    if backend_settings is None:
        backend_settings = getattr(settings, "WAGTAILFRONTENDCACHE", None)

    # Fallback to using WAGTAILFRONTENDCACHE_LOCATION setting (backwards compatibility)
    if backend_settings is None:
        cache_location = getattr(settings, "WAGTAILFRONTENDCACHE_LOCATION", None)

        if cache_location is not None:
            backend_settings = {
                "default": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.HTTPBackend",
                    "LOCATION": cache_location,
                },
            }

    # No settings found, return empty list
    if backend_settings is None:
        return {}

    backend_objects = {}

    for backend_name, _backend_config in backend_settings.items():
        if backends is not None and backend_name not in backends:
            continue

        backend_config = _backend_config.copy()
        backend = backend_config.pop("BACKEND")

        # Try to import the backend
        try:
            backend_cls = import_string(backend)
        except ImportError as e:
            raise InvalidFrontendCacheBackendError(
                f"Could not find backend '{backend}': {e}"
            )

        backend_objects[backend_name] = backend_cls(backend_config)

    return backend_objects


def purge_url_from_cache(url, backend_settings=None, backends=None):
    """
    Purge a single URL from the frontend cache.

    :param url: The URL to purge from the cache.
    :type url: str
    :param backend_settings: Optional custom backend settings to use instead of those defined in ``settings.WAGTAILFRONTENDCACHE``.
    :type backend_settings: dict, optional
    :param backends: Optional list of strings referencing specific backends from ``settings.WAGTAILFRONTENDCACHE`` or provided as ``backend_settings``. Can be used to limit purge operations to specific backends.
    :type backends: list, optional

    This function purges a single URL from the configured frontend cache backends. It's useful
    when you need to invalidate the cache for a specific URL.

    If no custom backends or settings are provided, it will use the default configuration
    from ``settings.WAGTAILFRONTENDCACHE``.

    NOTE: This function also handles internationalization, creating language-specific URLs if
    ``WAGTAILFRONTENDCACHE_LANGUAGES`` is set and ``USE_I18N`` is ``True``.
    """
    purge_urls_from_cache([url], backend_settings=backend_settings, backends=backends)


def purge_urls_from_cache(urls, backend_settings=None, backends=None):
    """
    Purge multiple URLs from the frontend cache.

    :param urls: An iterable of URLs to purge from the cache.
    :type urls: iterable of str
    :param backend_settings: Optional custom backend settings to use instead of those defined in ``settings.WAGTAILFRONTENDCACHE``.
    :type backend_settings: dict, optional
    :param backends: Optional list of strings referencing specific backends from ``settings.WAGTAILFRONTENDCACHE`` or provided as ``backend_settings``. Can be used to limit purge operations to specific backends.
    :type backends: list, optional

    This function purges multiple URLs from the configured frontend cache backends. It's useful
    when you need to invalidate the cache for multiple URLs at once.

    If no custom backends or settings are provided, it will use the default configuration
    from ``settings.WAGTAILFRONTENDCACHE``.

    NOTE: This function also handles internationalization, creating language-specific URLs if
    ``WAGTAILFRONTENDCACHE_LANGUAGES`` is set and ``USE_I18N`` is ``True``.
    """
    from .tasks import purge_urls_from_cache_task

    if not urls:
        return

    purge_urls_from_cache_task.enqueue(list(urls), backend_settings, backends)


def _get_page_cached_urls(page, cache_object=None):
    page_url = page.get_full_url(cache_object)
    if page_url is None:  # nothing to be done if the page has no routable URL
        return []

    return [
        page_url + path.lstrip("/")
        for path in page.specific_deferred.get_cached_paths()
    ]


def purge_page_from_cache(
    page, backend_settings=None, backends=None, *, cache_object=None
):
    """
    Purge a single page from the frontend cache.

    :param page: The page to purge from the cache.
    :type page: Page
    :param backend_settings: Optional custom backend settings to use instead of those defined in ``settings.WAGTAILFRONTENDCACHE``.
    :type backend_settings: dict, optional
    :param backends: Optional list of strings referencing specific backends from ``settings.WAGTAILFRONTENDCACHE`` or provided as ``backend_settings``. Can be used to limit purge operations to specific backends.
    :type backends: list, optional
    :param cache_object: Optional, but strongly recommended when making a series of requests to this method. An object to be passed to URL-related methods, allowing cached site root path data to be reused across multiple requests.
    :type cache_object: object, optional

    This function retrieves all cached URLs for the given page and purges them from the configured
    backends. It's useful when you need to invalidate the cache for a specific page,
    for example after the page has been updated.

    If no custom backends or settings are provided, it will use the default configuration
    from ``settings.WAGTAILFRONTENDCACHE``.

    The `cache_object` parameter can be any kind of object that supports arbitrary attribute
    assignment, such as a Python object or Django Model instance.
    """
    urls = _get_page_cached_urls(page, cache_object)
    purge_urls_from_cache(urls, backend_settings, backends)


def purge_pages_from_cache(
    pages, backend_settings=None, backends=None, *, cache_object=None
):
    """
    Purge multiple pages from the frontend cache.

    :param pages: An iterable of pages to purge from the cache.
    :type pages: iterable of Page
    :param backend_settings: Optional custom backend settings to use instead of those defined in ``settings.WAGTAILFRONTENDCACHE``.
    :type backend_settings: dict, optional
    :param backends: Optional list of strings matching keys from ``settings.WAGTAILFRONTENDCACHE`` or provided as ``backend_settings``. Can be used to limit purge operations to specific backends.
    :type backends: list, optional
    :param cache_object: Optional object to be passed to URL-related methods, to allow cached site root path data to be reused across multiple requests to this method. If not provided, the ``PurgeBatch`` object created by this method will be used instead.
    :type cache_object: object, optional

    This function retrieves all cached URLs for the given pages and purges them from the configured
    backends. It's useful when you need to invalidate the cache for multiple pages at once,
    for example after a bulk update operation.

    If no custom backends or settings are provided, it will use the default configuration
    from ``settings.WAGTAILFRONTENDCACHE``.

    The `cache_object` parameter can be any kind of object that supports arbitrary attribute
    assignment, such as a Python object or Django Model instance.
    """
    batch = PurgeBatch(cache_object=cache_object)
    batch.add_pages(pages)
    batch.purge(backend_settings, backends)


class PurgeBatch:
    """Represents a list of URLs to be purged in a single request"""

    def __init__(self, urls=None, *, cache_object=None):
        self.urls = set()

        if urls is not None:
            self.add_urls(urls)

        self.cache_object = cache_object

    def add_url(self, url):
        """Adds a single URL"""
        self.urls.add(url)

    def add_urls(self, urls):
        """
        Adds multiple URLs from an iterable

        This is equivalent to running ``.add_url(url)`` on each URL
        individually
        """
        self.urls.update(urls)

    def add_page(self, page):
        """
        Adds all URLs for the specified page

        This combines the page's full URL with each path that is returned by
        the page's `.get_cached_paths` method
        """
        self.add_urls(_get_page_cached_urls(page, self.cache_object or self))

    def add_pages(self, pages):
        """
        Adds multiple pages from a QuerySet or an iterable

        This is equivalent to running ``.add_page(page)`` on each page
        individually
        """
        for page in pages:
            self.add_page(page)

    def purge(self, backend_settings=None, backends=None):
        """
        Performs the purge of all the URLs in this batch

        This method takes two optional keyword arguments: backend_settings and backends

        - backend_settings can be used to override the WAGTAILFRONTENDCACHE setting for
          just this call

        - backends can be set to a list of backend names. When set, the invalidation request
          will only be sent to these backends
        """
        purge_urls_from_cache(self.urls, backend_settings, backends)



from django.conf import settings
from django.contrib.auth.models import Permission
from django.urls import reverse, reverse_lazy
from django.utils.functional import cached_property
from django.utils.http import urlencode
from django.utils.translation import gettext_lazy as _
from draftjs_exporter.dom import DOM

import wagtail.admin.rich_text.editors.draftail.features as draftail_features
from wagtail import __version__, hooks
from wagtail.admin.admin_url_finder import (
    ModelAdminURLFinder,
    register_admin_url_finder,
)
from wagtail.admin.auth import user_has_any_page_permission
from wagtail.admin.forms.collections import GroupCollectionManagementPermissionFormSet
from wagtail.admin.menu import (
    DismissibleMenuItem,
    DismissibleSubmenuMenuItem,
    MenuItem,
    SubmenuMenuItem,
    help_menu,
    reports_menu,
    settings_menu,
)
from wagtail.admin.rich_text.converters.contentstate import link_entity
from wagtail.admin.rich_text.converters.editor_html import (
    LinkTypeRule,
    PageLinkHandler,
    WhitelistRule,
)
from wagtail.admin.rich_text.converters.html_to_contentstate import (
    BlockElementHandler,
    ExternalLinkElementHandler,
    HorizontalRuleHandler,
    InlineStyleElementHandler,
    ListElementHandler,
    ListItemElementHandler,
    PageLinkElementHandler,
)
from wagtail.admin.search import SearchArea
from wagtail.admin.site_summary import PagesSummaryItem
from wagtail.admin.ui.menus.pages import PageMenuItem
from wagtail.admin.ui.sidebar import (
    PageExplorerMenuItem as PageExplorerMenuItemComponent,
)
from wagtail.admin.ui.sidebar import SubMenuItem as SubMenuItemComponent
from wagtail.admin.views.pages.bulk_actions import (
    DeleteBulkAction,
    MoveBulkAction,
    PublishBulkAction,
    UnpublishBulkAction,
)
from wagtail.admin.viewsets import viewsets
from wagtail.admin.widgets import ButtonWithDropdownFromHook
from wagtail.models import Collection, Page, Task, Workflow
from wagtail.permissions import (
    collection_permission_policy,
    page_permission_policy,
    task_permission_policy,
    workflow_permission_policy,
)
from wagtail.templatetags.wagtailcore_tags import (
    wagtail_feature_release_editor_guide_link,
    wagtail_feature_release_whats_new_link,
)
from wagtail.utils.version import get_main_version
from wagtail.whitelist import allow_without_attributes, attribute_rule, check_url


class ExplorerMenuItem(MenuItem):
    def is_shown(self, request):
        return user_has_any_page_permission(request.user)

    def get_context(self, request):
        context = super().get_context(request)
        start_page = page_permission_policy.explorable_root_instance(request.user)

        if start_page:
            context["start_page_id"] = start_page.id

        return context

    def render_component(self, request):
        start_page = page_permission_policy.explorable_root_instance(request.user)

        if start_page:
            return PageExplorerMenuItemComponent(
                self.name,
                self.label,
                self.url,
                start_page.id,
                icon_name=self.icon_name,
                classname=self.classname,
            )
        else:
            return super().render_component(request)


@hooks.register("register_admin_menu_item")
def register_explorer_menu_item():
    return ExplorerMenuItem(
        _("Pages"),
        reverse("wagtailadmin_explore_root"),
        name="explorer",
        icon_name="folder-open-inverse",
        order=100,
    )


class SettingsMenuItem(SubmenuMenuItem):
    def render_component(self, request):
        return SubMenuItemComponent(
            self.name,
            self.label,
            self.menu.render_component(request),
            icon_name=self.icon_name,
            classname=self.classname,
            footer_text="Wagtail v" + __version__,
        )


@hooks.register("register_admin_menu_item")
def register_settings_menu():
    return SettingsMenuItem(
        _("Settings"),
        settings_menu,
        name="settings",
        icon_name="cogs",
        order=10000,
    )


@hooks.register("register_permissions")
def register_permissions():
    return Permission.objects.filter(
        content_type__app_label="wagtailadmin", codename="access_admin"
    )


class PageSearchArea(SearchArea):
    def __init__(self):
        super().__init__(
            _("Pages"),
            reverse("wagtailadmin_pages:search"),
            name="pages",
            icon_name="folder-open-inverse",
            order=100,
        )

    def is_shown(self, request):
        return user_has_any_page_permission(request.user)


@hooks.register("register_admin_search_area")
def register_pages_search_area():
    return PageSearchArea()


@hooks.register("register_group_permission_panel")
def register_collection_permissions_panel():
    return GroupCollectionManagementPermissionFormSet


class CollectionsMenuItem(MenuItem):
    def is_shown(self, request):
        return collection_permission_policy.user_has_any_permission(
            request.user, ["add", "change", "delete"]
        )


@hooks.register("register_settings_menu_item")
def register_collections_menu_item():
    return CollectionsMenuItem(
        _("Collections"),
        reverse("wagtailadmin_collections:index"),
        name="collections",
        icon_name="folder-open-1",
        order=700,
    )


class WorkflowsMenuItem(MenuItem):
    def is_shown(self, request):
        if not getattr(settings, "WAGTAIL_WORKFLOW_ENABLED", True):
            return False

        return workflow_permission_policy.user_has_any_permission(
            request.user, ["add", "change", "delete"]
        )


class WorkflowTasksMenuItem(MenuItem):
    def is_shown(self, request):
        if not getattr(settings, "WAGTAIL_WORKFLOW_ENABLED", True):
            return False

        return task_permission_policy.user_has_any_permission(
            request.user, ["add", "change", "delete"]
        )


@hooks.register("register_settings_menu_item")
def register_workflows_menu_item():
    return WorkflowsMenuItem(
        _("Workflows"),
        reverse("wagtailadmin_workflows:index"),
        name="workflows",
        icon_name="tasks",
        order=100,
    )


@hooks.register("register_settings_menu_item")
def register_workflow_tasks_menu_item():
    return WorkflowTasksMenuItem(
        _("Workflow tasks"),
        reverse("wagtailadmin_workflows:task_index"),
        name="workflow-tasks",
        icon_name="thumbtack",
        order=150,
    )


@hooks.register("register_page_listing_buttons")
def page_listing_buttons(page, user, next_url=None):
    yield ButtonWithDropdownFromHook(
        "",
        hook_name="register_page_listing_more_buttons",
        page=page,
        user=user,
        next_url=next_url,
        icon_name="dots-horizontal",
        attrs={
            "aria-label": _("More options for '%(title)s'")
            % {"title": page.get_admin_display_title()},
        },
        priority=50,
    )


class PageListingEditButton(PageMenuItem):
    label = _("Edit")
    icon_name = "edit"
    url_name = "wagtailadmin_pages:edit"

    def is_shown(self, user):
        return self.page.permissions_for_user(user).can_edit()


class PageListingViewDraftButton(PageMenuItem):
    label = _("View draft")
    icon_name = "draft"
    url_name = "wagtailadmin_pages:view_draft"
    link_rel = "noreferrer"

    def is_shown(self, user):
        return self.page.has_unpublished_changes and self.page.is_previewable()


class PageListingViewLiveButton(PageMenuItem):
    label = _("View live")
    icon_name = "doc-empty"
    link_rel = "noreferrer"

    def is_shown(self, user):
        return self.page.live and self.page.url


class PageListingAddChildPageButton(PageMenuItem):
    label = _("Add child page")
    icon_name = "circle-plus"
    url_name = "wagtailadmin_pages:add_subpage"

    def is_shown(self, user):
        return self.page.permissions_for_user(user).can_add_subpage()


class PageListingMoveButton(PageMenuItem):
    label = _("Move")
    icon_name = "arrow-right-full"
    url_name = "wagtailadmin_pages:move"

    def is_shown(self, user):
        return self.page.permissions_for_user(user).can_move()


class PageListingCopyButton(PageMenuItem):
    label = _("Copy")
    icon_name = "copy"
    url_name = "wagtailadmin_pages:copy"

    def is_shown(self, user):
        return self.page.permissions_for_user(user).can_copy()


class PageListingDeleteButton(PageMenuItem):
    label = _("Delete")
    icon_name = "bin"

    @cached_property
    def url(self):
        if self.page:
            url = reverse("wagtailadmin_pages:delete", args=[self.page.id])
            if self.next_url:
                if self.next_url == reverse(
                    "wagtailadmin_explore", args=[self.page.id]
                ):
                    # cannot redirect to the explore view after deleting the page
                    pass
                elif self.next_url == reverse(
                    "wagtailadmin_pages:edit", args=[self.page.id]
                ):
                    # cannot redirect to the edit view after deleting the page
                    pass
                else:
                    # OK to add the 'next' parameter
                    url += "?" + urlencode({"next": self.next_url})
            return url

    def is_shown(self, user):
        return self.page.permissions_for_user(user).can_delete()


class PageListingUnpublishButton(PageMenuItem):
    label = _("Unpublish")
    icon_name = "download"
    url_name = "wagtailadmin_pages:unpublish"

    def is_shown(self, user):
        return self.page.permissions_for_user(user).can_unpublish()


class PageListingHistoryButton(PageMenuItem):
    label = _("History")
    icon_name = "history"
    url_name = "wagtailadmin_pages:history"

    def is_shown(self, user):
        return self.page.permissions_for_user(user).can_view_revisions()


class PageListingSortMenuOrderButton(PageMenuItem):
    label = _("Sort menu order")
    icon_name = "list-ul"

    def is_shown(self, user):
        return self.page.permissions_for_user(user).can_reorder_children()

    @cached_property
    def url(self):
        return reverse("wagtailadmin_explore", args=[self.page.id]) + "?ordering=ord"


@hooks.register("register_page_listing_more_buttons")
def page_listing_more_buttons(page, user, next_url=None):
    yield PageListingEditButton(page=page, next_url=next_url, priority=2)
    yield PageListingViewDraftButton(page=page, priority=4)
    yield PageListingViewLiveButton(page=page, url=page.url, priority=6)
    yield PageListingAddChildPageButton(page=page, next_url=next_url, priority=8)
    yield PageListingMoveButton(page=page, priority=10)
    yield PageListingCopyButton(page=page, next_url=next_url, priority=20)
    yield PageListingDeleteButton(page=page, next_url=next_url, priority=30)
    yield PageListingUnpublishButton(page=page, next_url=next_url, priority=40)
    yield PageListingHistoryButton(page=page, priority=50)
    yield PageListingSortMenuOrderButton(page=page, priority=60)


@hooks.register("register_page_header_buttons")
def page_header_buttons(page, user, view_name, next_url=None):
    yield PageListingEditButton(page=page, priority=10)

    # "add child" is a separate primary action on the index page
    if view_name != "index":
        yield PageListingAddChildPageButton(page=page, priority=15)

    yield PageListingMoveButton(page=page, priority=20)
    yield PageListingCopyButton(page=page, next_url=next_url, priority=30)
    yield PageListingDeleteButton(page=page, next_url=next_url, priority=50)
    yield PageListingUnpublishButton(page=page, next_url=next_url, priority=60)
    yield PageListingHistoryButton(page=page, priority=65)
    yield PageListingSortMenuOrderButton(page=page, priority=70)


@hooks.register("register_admin_urls")
def register_viewsets_urls():
    viewsets.populate()
    return viewsets.get_urlpatterns()


@hooks.register("register_rich_text_features")
def register_core_features(features):
    features.register_converter_rule(
        "editorhtml",
        "link",
        [
            WhitelistRule("a", attribute_rule({"href": check_url})),
            LinkTypeRule("page", PageLinkHandler),
        ],
    )

    features.register_converter_rule(
        "editorhtml",
        "bold",
        [
            WhitelistRule("b", allow_without_attributes),
            WhitelistRule("strong", allow_without_attributes),
        ],
    )

    features.register_converter_rule(
        "editorhtml",
        "italic",
        [
            WhitelistRule("i", allow_without_attributes),
            WhitelistRule("em", allow_without_attributes),
        ],
    )

    headings_elements = ["h1", "h2", "h3", "h4", "h5", "h6"]
    for order, element in enumerate(headings_elements):
        features.register_converter_rule(
            "editorhtml", element, [WhitelistRule(element, allow_without_attributes)]
        )

    features.register_converter_rule(
        "editorhtml",
        "ol",
        [
            WhitelistRule("ol", allow_without_attributes),
            WhitelistRule("li", allow_without_attributes),
        ],
    )

    features.register_converter_rule(
        "editorhtml",
        "ul",
        [
            WhitelistRule("ul", allow_without_attributes),
            WhitelistRule("li", allow_without_attributes),
        ],
    )

    # Draftail
    features.register_editor_plugin(
        "draftail", "hr", draftail_features.BooleanFeature("enableHorizontalRule")
    )
    features.register_converter_rule(
        "contentstate",
        "hr",
        {
            "from_database_format": {
                "hr": HorizontalRuleHandler(),
            },
            "to_database_format": {
                "entity_decorators": {
                    "HORIZONTAL_RULE": lambda props: DOM.create_element("hr")
                }
            },
        },
    )

    features.register_editor_plugin(
        "draftail",
        "h1",
        draftail_features.BlockFeature(
            {
                "icon": "h1",
                "type": "header-one",
                "description": _("Heading 1"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "h1",
        {
            "from_database_format": {
                "h1": BlockElementHandler("header-one"),
            },
            "to_database_format": {"block_map": {"header-one": "h1"}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "h2",
        draftail_features.BlockFeature(
            {
                "icon": "h2",
                "type": "header-two",
                "description": _("Heading 2"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "h2",
        {
            "from_database_format": {
                "h2": BlockElementHandler("header-two"),
            },
            "to_database_format": {"block_map": {"header-two": "h2"}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "h3",
        draftail_features.BlockFeature(
            {
                "icon": "h3",
                "type": "header-three",
                "description": _("Heading 3"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "h3",
        {
            "from_database_format": {
                "h3": BlockElementHandler("header-three"),
            },
            "to_database_format": {"block_map": {"header-three": "h3"}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "h4",
        draftail_features.BlockFeature(
            {
                "icon": "h4",
                "type": "header-four",
                "description": _("Heading 4"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "h4",
        {
            "from_database_format": {
                "h4": BlockElementHandler("header-four"),
            },
            "to_database_format": {"block_map": {"header-four": "h4"}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "h5",
        draftail_features.BlockFeature(
            {
                "icon": "h5",
                "type": "header-five",
                "description": _("Heading 5"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "h5",
        {
            "from_database_format": {
                "h5": BlockElementHandler("header-five"),
            },
            "to_database_format": {"block_map": {"header-five": "h5"}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "h6",
        draftail_features.BlockFeature(
            {
                "icon": "h6",
                "type": "header-six",
                "description": _("Heading 6"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "h6",
        {
            "from_database_format": {
                "h6": BlockElementHandler("header-six"),
            },
            "to_database_format": {"block_map": {"header-six": "h6"}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "ul",
        draftail_features.BlockFeature(
            {
                "type": "unordered-list-item",
                "icon": "list-ul",
                "description": _("Bulleted list"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "ul",
        {
            "from_database_format": {
                "ul": ListElementHandler("unordered-list-item"),
                "li": ListItemElementHandler(),
            },
            "to_database_format": {
                "block_map": {"unordered-list-item": {"element": "li", "wrapper": "ul"}}
            },
        },
    )
    features.register_editor_plugin(
        "draftail",
        "ol",
        draftail_features.BlockFeature(
            {
                "type": "ordered-list-item",
                "icon": "list-ol",
                "description": _("Numbered list"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "ol",
        {
            "from_database_format": {
                "ol": ListElementHandler("ordered-list-item"),
                "li": ListItemElementHandler(),
            },
            "to_database_format": {
                "block_map": {"ordered-list-item": {"element": "li", "wrapper": "ol"}}
            },
        },
    )
    features.register_editor_plugin(
        "draftail",
        "blockquote",
        draftail_features.BlockFeature(
            {
                "type": "blockquote",
                "icon": "openquote",
                "description": _("Blockquote"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "blockquote",
        {
            "from_database_format": {
                "blockquote": BlockElementHandler("blockquote"),
            },
            "to_database_format": {"block_map": {"blockquote": "blockquote"}},
        },
    )

    features.register_editor_plugin(
        "draftail",
        "bold",
        draftail_features.InlineStyleFeature(
            {
                "type": "BOLD",
                "icon": "bold",
                "description": _("Bold"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "bold",
        {
            "from_database_format": {
                "b": InlineStyleElementHandler("BOLD"),
                "strong": InlineStyleElementHandler("BOLD"),
            },
            "to_database_format": {"style_map": {"BOLD": "b"}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "italic",
        draftail_features.InlineStyleFeature(
            {
                "type": "ITALIC",
                "icon": "italic",
                "description": _("Italic"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "italic",
        {
            "from_database_format": {
                "i": InlineStyleElementHandler("ITALIC"),
                "em": InlineStyleElementHandler("ITALIC"),
            },
            "to_database_format": {"style_map": {"ITALIC": "i"}},
        },
    )

    features.register_editor_plugin(
        "draftail",
        "link",
        draftail_features.EntityFeature(
            {
                "type": "LINK",
                "icon": "link",
                "description": _("Link"),
                # We want to enforce constraints on which links can be pasted into rich text.
                # Keep only the attributes Wagtail needs.
                "attributes": ["url", "id", "parentId"],
                "allowlist": {
                    # Keep pasted links with http/https protocol, and not-pasted links (href = undefined).
                    "href": "^(http:|https:|undefined$)",
                },
                "chooserUrls": {
                    "pageChooser": reverse_lazy("wagtailadmin_choose_page"),
                    "externalLinkChooser": reverse_lazy(
                        "wagtailadmin_choose_page_external_link"
                    ),
                    "emailLinkChooser": reverse_lazy(
                        "wagtailadmin_choose_page_email_link"
                    ),
                    "phoneLinkChooser": reverse_lazy(
                        "wagtailadmin_choose_page_phone_link"
                    ),
                    "anchorLinkChooser": reverse_lazy(
                        "wagtailadmin_choose_page_anchor_link"
                    ),
                },
            },
            js=[
                "wagtailadmin/js/page-chooser-modal.js",
            ],
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "link",
        {
            "from_database_format": {
                "a[href]": ExternalLinkElementHandler("LINK"),
                'a[linktype="page"]': PageLinkElementHandler("LINK"),
            },
            "to_database_format": {"entity_decorators": {"LINK": link_entity}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "superscript",
        draftail_features.InlineStyleFeature(
            {
                "type": "SUPERSCRIPT",
                "icon": "superscript",
                "description": _("Superscript"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "superscript",
        {
            "from_database_format": {
                "sup": InlineStyleElementHandler("SUPERSCRIPT"),
            },
            "to_database_format": {"style_map": {"SUPERSCRIPT": "sup"}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "subscript",
        draftail_features.InlineStyleFeature(
            {
                "type": "SUBSCRIPT",
                "icon": "subscript",
                "description": _("Subscript"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "subscript",
        {
            "from_database_format": {
                "sub": InlineStyleElementHandler("SUBSCRIPT"),
            },
            "to_database_format": {"style_map": {"SUBSCRIPT": "sub"}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "strikethrough",
        draftail_features.InlineStyleFeature(
            {
                "type": "STRIKETHROUGH",
                "icon": "strikethrough",
                "description": _("Strikethrough"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "strikethrough",
        {
            "from_database_format": {
                "s": InlineStyleElementHandler("STRIKETHROUGH"),
            },
            "to_database_format": {"style_map": {"STRIKETHROUGH": "s"}},
        },
    )
    features.register_editor_plugin(
        "draftail",
        "code",
        draftail_features.InlineStyleFeature(
            {
                "type": "CODE",
                "icon": "code",
                "description": _("Code"),
            }
        ),
    )
    features.register_converter_rule(
        "contentstate",
        "code",
        {
            "from_database_format": {
                "code": InlineStyleElementHandler("CODE"),
            },
            "to_database_format": {"style_map": {"CODE": "code"}},
        },
    )


class LockedPagesMenuItem(MenuItem):
    def is_shown(self, request):
        return page_permission_policy.user_has_permission(request.user, "unlock")


class WorkflowReportMenuItem(MenuItem):
    def is_shown(self, request):
        return getattr(
            settings, "WAGTAIL_WORKFLOW_ENABLED", True
        ) and page_permission_policy.user_has_any_permission(
            request.user, ["add", "change", "publish"]
        )


class SiteHistoryReportMenuItem(MenuItem):
    def is_shown(self, request):
        return page_permission_policy.explorable_root_instance(request.user) is not None


class AgingPagesReportMenuItem(MenuItem):
    def is_shown(self, request):
        return getattr(
            settings, "WAGTAIL_AGING_PAGES_ENABLED", True
        ) and page_permission_policy.user_has_any_permission(
            request.user, ["add", "change", "publish"]
        )


class PageTypesReportMenuItem(MenuItem):
    def is_shown(self, request):
        return page_permission_policy.user_has_any_permission(
            request.user, ["add", "change", "publish"]
        )


@hooks.register("register_reports_menu_item")
def register_locked_pages_menu_item():
    return LockedPagesMenuItem(
        _("Locked pages"),
        reverse("wagtailadmin_reports:locked_pages"),
        name="locked-pages",
        icon_name="lock",
        order=700,
    )


@hooks.register("register_reports_menu_item")
def register_workflow_report_menu_item():
    return WorkflowReportMenuItem(
        _("Workflows"),
        reverse("wagtailadmin_reports:workflow"),
        name="workflows",
        icon_name="tasks",
        order=800,
    )


@hooks.register("register_reports_menu_item")
def register_workflow_tasks_report_menu_item():
    return WorkflowReportMenuItem(
        _("Workflow tasks"),
        reverse("wagtailadmin_reports:workflow_tasks"),
        name="workflow-tasks",
        icon_name="thumbtack",
        order=900,
    )


@hooks.register("register_reports_menu_item")
def register_site_history_report_menu_item():
    return SiteHistoryReportMenuItem(
        _("Site history"),
        reverse("wagtailadmin_reports:site_history"),
        name="site-history",
        icon_name="history",
        order=1000,
    )


@hooks.register("register_reports_menu_item")
def register_aging_pages_report_menu_item():
    return AgingPagesReportMenuItem(
        _("Aging pages"),
        reverse("wagtailadmin_reports:aging_pages"),
        name="aging-pages",
        icon_name="time",
        order=1100,
    )


@hooks.register("register_reports_menu_item")
def register_page_types_report_menu_item():
    return PageTypesReportMenuItem(
        _("Page types usage"),
        reverse("wagtailadmin_reports:page_types_usage"),
        name="page-types-usage",
        icon_name="doc-empty-inverse",
        order=1200,
    )


@hooks.register("register_admin_menu_item")
def register_reports_menu():
    return SubmenuMenuItem(
        _("Reports"),
        reports_menu,
        name="reports",
        icon_name="site",
        order=9000,
    )


@hooks.register("register_help_menu_item")
def register_whats_new_in_wagtail_version_menu_item():
    version = get_main_version(include_patch=False)
    return DismissibleMenuItem(
        _("What's new in Wagtail %(version)s") % {"version": version},
        wagtail_feature_release_whats_new_link(),
        icon_name="help",
        order=1000,
        attrs={"target": "_blank", "rel": "noreferrer"},
        name=f"whats-new-in-wagtail-{version}",
    )


@hooks.register("register_help_menu_item")
def register_editors_guide_menu_item():
    return DismissibleMenuItem(
        _("Editor Guide"),
        wagtail_feature_release_editor_guide_link(),
        icon_name="help",
        order=1100,
        attrs={"target": "_blank", "rel": "noreferrer"},
        name="editor-guide",
    )


@hooks.register("register_help_menu_item")
def register_keyboard_shortcuts_menu_item():
    """
    Triggers the keyboard shortcuts dialog to open when clicked
    while preventing the default link click action.
    """

    return MenuItem(
        _("Shortcuts"),
        icon_name="keyboard",
        order=1200,
        attrs={
            "role": "button",  # Ensure screen readers announce this as a button
            "data-a11y-dialog-show": "keyboard-shortcuts-dialog",
            "data-action": "w-action#noop:prevent:stop",
            "data-controller": "w-action",
        },
        name="keyboard-shortcuts-trigger",
        url="#",
    )


@hooks.register("register_admin_menu_item")
def register_help_menu():
    return DismissibleSubmenuMenuItem(
        _("Help"),
        help_menu,
        name="help",
        icon_name="help",
        order=11000,
    )


@hooks.register("register_icons")
def register_icons(icons):
    for icon in [
        "arrow-down.svg",
        "arrow-right-full.svg",
        "arrow-left.svg",
        "arrow-right.svg",
        "arrow-up.svg",
        "bars.svg",
        "bin.svg",
        "bold.svg",
        "breadcrumb-expand.svg",
        "calendar.svg",
        "calendar-alt.svg",
        "calendar-check.svg",
        "check.svg",
        "circle-check.svg",
        "circle-plus.svg",
        "circle-xmark.svg",
        "clipboard-list.svg",
        "code.svg",
        "cog.svg",
        "cogs.svg",
        "copy.svg",
        "collapse-down.svg",
        "collapse-up.svg",
        "comment.svg",
        "comment-add.svg",
        "comment-add-reversed.svg",
        "cross.svg",
        "crosshairs.svg",
        "cut.svg",
        "date.svg",
        "decimal.svg",
        "desktop.svg",
        "doc-empty-inverse.svg",
        "doc-empty.svg",
        "doc-full-inverse.svg",
        "doc-full.svg",
        "dots-horizontal.svg",
        "download.svg",
        "draft.svg",
        "edit.svg",
        "expand-right.svg",
        "error.svg",
        "folder-inverse.svg",
        "folder-open-1.svg",
        "folder-open-inverse.svg",
        "folder.svg",
        "form.svg",
        "glasses.svg",
        "globe.svg",
        "grid.svg",
        "grip.svg",
        "group.svg",
        "h1.svg",
        "h2.svg",
        "h3.svg",
        "h4.svg",
        "h5.svg",
        "h6.svg",
        "help.svg",
        "history.svg",
        "home.svg",
        "image.svg",
        "info-circle.svg",
        "italic.svg",
        "key.svg",
        "keyboard.svg",
        "link.svg",
        "link-external.svg",
        "list-ol.svg",
        "list-ul.svg",
        "lock-open.svg",
        "lock.svg",
        "login.svg",
        "logout.svg",
        "mail.svg",
        "media.svg",
        "minus.svg",
        "mobile-alt.svg",
        "no-view.svg",
        "openquote.svg",
        "order-down.svg",
        "order-up.svg",
        "order.svg",
        "password.svg",
        "pick.svg",
        "pilcrow.svg",
        "placeholder.svg",
        "plus-inverse.svg",
        "plus.svg",
        "radio-empty.svg",
        "radio-full.svg",
        "redirect.svg",
        "regex.svg",
        "resubmit.svg",
        "rotate.svg",
        "search.svg",
        "site.svg",
        "sliders.svg",
        "snippet.svg",
        "spinner.svg",
        "strikethrough.svg",
        "success.svg",
        "subscript.svg",
        "superscript.svg",
        "table.svg",
        "tablet-alt.svg",
        "tag.svg",
        "tasks.svg",
        "thumbtack.svg",
        "thumbtack-crossed.svg",
        "tick-inverse.svg",
        "time.svg",
        "title.svg",
        "upload.svg",
        "user.svg",
        "view.svg",
        "wagtail.svg",
        "warning.svg",
    ]:
        icons.append(f"wagtailadmin/icons/{icon}")
    return icons


@hooks.register("construct_homepage_summary_items")
def add_pages_summary_item(request, items):
    items.insert(0, PagesSummaryItem(request))


class PageAdminURLFinder:
    def __init__(self, user):
        self.user = user

    def get_edit_url(self, instance):
        if self.user and not instance.permissions_for_user(self.user).can_edit():
            return None
        else:
            return reverse("wagtailadmin_pages:edit", args=(instance.pk,))


register_admin_url_finder(Page, PageAdminURLFinder)


class CollectionAdminURLFinder(ModelAdminURLFinder):
    permission_policy = collection_permission_policy
    edit_url_name = "wagtailadmin_collections:edit"


register_admin_url_finder(Collection, CollectionAdminURLFinder)


class WorkflowAdminURLFinder(ModelAdminURLFinder):
    permission_policy = workflow_permission_policy
    edit_url_name = "wagtailadmin_workflows:edit"


register_admin_url_finder(Workflow, WorkflowAdminURLFinder)


class WorkflowTaskAdminURLFinder(ModelAdminURLFinder):
    permission_policy = task_permission_policy
    edit_url_name = "wagtailadmin_workflows:edit_task"


register_admin_url_finder(Task, WorkflowTaskAdminURLFinder)


for action_class in [
    DeleteBulkAction,
    MoveBulkAction,
    PublishBulkAction,
    UnpublishBulkAction,
]:
    hooks.register("register_bulk_action", action_class)
